//=== DIRCTIVES ===============================================================

#if defined(_WIN32) || defined(WIN32)
#if defined(_MSC_VER)
#pragma warning(disable : 4244)		// 'short' to 'uchar' conversion warning
#pragma warning(disable : 4786)		// long identifier warning
#endif
#endif


//================= INCLUDES ====================================================

#include <vector>
#include <iostream>
#include <algorithm>
#include <functional>
#include <FL/fl_ask.h>


#include "PDecompRenderer.h"
#include "../Core_Mesh/Iterators.h"
#include "../Core_Mesh/MeshDelegate.h"

#include "../Utility/GetNormals.h"
#include "../Utility/GetUV.h"
#include "../Core_Mesh/TraitParser.h"
#include "../Core_Mesh/MeshUtility.h"

//#include "../Utility/XUtility.h"

using namespace XMeshLib;
using namespace std;

//== IMPLEMENTATION ===========================================================

const float PDecompRenderer::sharpFaceColor[] = {0.0f, 0.4f, 0.9f};
const float PDecompRenderer::sharpVerColor[] = {1.0f, 1.0f, 0.0f};
const float PDecompRenderer::sharpEdgeColor[] = {1.0f, 1.0f, 0.0f};
const float PDecompRenderer::surfaceColor[] = {1.0f, 1.0f, 0.6f};
const float PDecompRenderer::normalColor[] = {1.0f,1.0f,1.0f};
const float PDecompRenderer::edgeColor[] = {0.0f,0.0f,0.0f};
const float PDecompRenderer::pointColor[] = {0.7f, 0.7f, 0.7f};
	
const float PDecompRenderer::HomoRGB[][3] = {
	1,0,0, 
	0,1,0, 
	0,0,1, 
	1,1,0, 
	1,0,1, 
	0,1,1, 
	0.72f,0.53f,0.04f, 
	0,0,0,  
	0.75f, 0.75f, 0.75f};


PDecompRenderer::PDecompRenderer(int x, int y, int w, int h, const char *label):
Fl_Gl_Window(x, y, w, h, label), renderingMesh(NULL)
{
	Init();	
}

PDecompRenderer::~PDecompRenderer()
{
	if (m_recording)
		EndVideoCapture();
	for (int i=0;i<guideEdges.size();++i)
	{
		Curve * c= guideEdges[i];
		if (c)
			delete c;		
	}
	guideEdges.clear();
}


int PDecompRenderer::handle(int event)
{
	int button;
	int x, y;
	switch(event) {
	case FL_PUSH:
		button = Fl::event_button();
		x = Fl::event_x();
		y = Fl::event_y();
		MouseDown(button, x, y);
		return 1;
	case FL_DRAG:
		button = Fl::event_button();
		x = Fl::event_x();
		y = Fl::event_y();
		MouseMotion(button, x, y);
		return 1;
	case FL_RELEASE:   
		button = Fl::event_button();
		x = Fl::event_x();
		y = Fl::event_y();
		MouseUp(button, x, y);
		return 1;
	case FL_FOCUS :
	case FL_UNFOCUS :
//    ... Return 1 if you want keyboard events, 0 otherwise
    return 1;
	case FL_KEYBOARD:
//    ... keypress, key is in Fl::event_key(), ascii in Fl::event_text()
//    ... Return 1 if you understand/use the keyboard event, 0 otherwise...
	case FL_SHORTCUT:
//   ... shortcut, key is in Fl::event_key(), ascii in Fl::event_text()
//    ... Return 1 if you understand/use the shortcut event, 0 otherwise...
    return 1;
	default:
    // pass other events to the base class...
	return Fl_Gl_Window::handle(event);
  }
}

void PDecompRenderer::draw()
{
	// Fl_Gl_Window::valid() is turned off when FLTK creates a new context 
	// for this window or when the window resizes, and is turned on after draw() is called.
	if (!initialized)
	{
		InitGL();
	
		GLenum err = glewInit();
		if (GLEW_OK != err)
		{
			/* Problem: glewInit failed, something is seriously wrong. */
			fprintf(stderr, "Error: %s\n", glewGetErrorString(err));
			return;
		}
		fprintf(stdout, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));

		initialized = true;
	}

	RenderGL();

	if (m_recording)
	{
		//assert(m_avigen);
		//CaptureVideoFrame();
	}
}


// Mouse event helper function - rotate
void PDecompRenderer::Rotate(float ox, float oy, float nx, float ny)
{
	// Random multiplicative factor - adjust to taste
//	nx += (ROT_MULT-1)*(nx-ox);
//	ny += (ROT_MULT-1)*(ny-oy);

	quat q;
	Camera::Arc2Quaternion(ox, oy, nx, ny, q);

	if (rotate_dist == 0) {
		q[1] = -q[1];
		q[2] = -q[2];
	}

	theCamera.Rotate(q, rotate_dist);
}


// Mouse event helper function - translate
void PDecompRenderer::Translate(float dx, float dy, float dz)
{
	// some auto param ajustment to be refined later
	float scalefactor = 0.5f*view_dist;
	dx *= scalefactor*theCamera.fov;
	dy *= scalefactor*theCamera.fov;
	dz = view_dist * (exp(-0.5f * dz) - 1.0f);

	// moving the camera in the opposite direction to
	// the direction we want the object to move
	theCamera.Move(-dx, -dy, -dz);
	UpdateDist(-dx, -dy, -dz);
}


// Mouse event helper function - update distance
void PDecompRenderer::UpdateDist(float dx, float dy, float dz)
{
	vec3 viewdir(0, 0, -view_dist);
	vec3 delta(-dx, -dy, -dz);

	vec3 newviewdir = viewdir + delta;

	view_dist = -newviewdir.z;
	// Always rotate around the center of the object
	// which is assumed to be place at the origin of the world coordinate
	rotate_dist = view_dist;

}


// Helper function - convert integer coordinates of mouse position to
// normalized floats
void PDecompRenderer::Mousei2f(int x, int y, float *xf, float *yf)
{
	int screenwidth = Screenw();
	int screenheight = Screenh();

	float r = (float)screenwidth/screenheight;
	if (r > 1.0) {
		*xf = r * (2.0f * x / screenwidth - 1.0f);
		*yf = 1.0f - 2.0f * y / screenheight;
	} else {
		*xf = 2.0f * x / screenwidth - 1.0f;
		*yf = (1.0f - 2.0f * y / screenheight) / r;
	}

}

// Reshape function
void PDecompRenderer::Reshape(GLsizei w, GLsizei h)
{
	glViewport(0, 0, w, h);

	win_width = w;
	win_height = h;
}


// Initialize the renderer state machine
void PDecompRenderer::Init()
{
	initialized = false;

	rotate_dist = 0;
	view_dist = 0;

	shading_mode = SHADING_SMOOTH;
	material_mode = MATERIAL_DEFAULT;
	polygon_mode = POLYGON_FILL;

	enable_lighting = 1;
	enable_picking = 0;
	picking_mode = PICKING_VERTEX;

	show_normal = 0;
	show_boundingbox = 0;
	show_triangulation = 0;
	show_homo = 1;
	show_sphere = 0;
	show_pset = 0;
	show_hexagons = 0;
	show_hexagonUV = 0;
	show_morph = 0;
	show_KMRGB = show_CFRGB = 0;
	show_psetboundary = 0;
	show_psetseparate = 0;
	show_psetbmarker = 0;
	show_pantInd = 0;
	show_psetrect = 0;
	show_sharpvert = 1;
	show_sharpedge = 1;
	show_sharpface = 1;
	show_texcoord = 0;	
	show_envmap = 0;
	show_floor = 0;
	show_source = 0;
	show_collocation = 0;
	show_stats = 1;
	show_markers = 1;
	show_pantsfeatures = 0;


	gl_version = NULL;
	gl_vendor = NULL;
	gl_renderer = NULL;
	gl_extensions = NULL;
	glu_extensions = NULL;

	win_width = 0;
	win_height = 0;
	fullscreen = false;

	//R = 0.6f; G= 0.6f; B = 0.6f;

	for (int i = 0; i < 256; ++i)
		key_down[i] = 0;

	mesh.resize(2, NULL); //at least one for pd.mesh, one for pd.basePatch
	
	pd = NULL;
	psMap = NULL;
	pickHomoInd1 = pickHomoInd2 = 0;

	surgeryHandle = -1;
	
	//m_vert_rgb_ok = false;
	//m_vert_opos_ok = false;
	//m_vert_onormal_ok = false;
	//m_vert_orgb_ok = false;
	//m_blendc = 0;

	//m_avigen = NULL;
	m_recording = false;

	cMeshInd = 0;

	colorCode_type = 0;
}


void PDecompRenderer::InitGL()
{
	gl_version = glGetString(GL_VERSION);
	gl_vendor = glGetString(GL_VENDOR);
	gl_renderer = glGetString(GL_RENDERER);
	gl_extensions = glGetString(GL_EXTENSIONS);
	glu_extensions = gluGetString(GLU_EXTENSIONS);

	glClearColor(R, G, B, 0.0f);
	glPolygonMode(GL_FRONT, GL_FILL);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_DEPTH_TEST);

	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);
	//glEnable(GL_CULL_FACE);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	// set pixel alignment
	glPixelStorei(GL_PACK_ALIGNMENT, 1);

	InitLighting();
	InitMaterial();
	InitQuadrics();
	//InitTextureData();	
	InitCamera();
}

void PDecompRenderer::InitLighting()
{
	GLfloat	light0_Ka[] = {0.3f,0.3f,0.3f,1.0f};
	GLfloat	light0_Kd[] = {0.8f,0.8f,0.8f,1.0f};
	GLfloat light0_Ks[]= {0.8f,0.8f,0.8f,1.0f};
	
	GLfloat light0_Pos[] = {0.0f,5.0f,5.0f,0.0f};
	GLfloat light1_Pos[] = {0.0f,-5.0f,6.0f,0.0f};
	GLfloat light2_Pos[] = {0.0f,0.0f,10.0f,0.0f};
	GLfloat light3_Pos[] = {0.0f,0.0f,-10.0f,0.0f};

	glPushMatrix();

	glLightfv(GL_LIGHT0, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT0, GL_POSITION, light0_Pos);
	glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT1, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT1, GL_POSITION, light1_Pos);
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT2, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT2, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT2, GL_POSITION, light2_Pos);
	glLightf(GL_LIGHT2, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT2, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT2, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightfv(GL_LIGHT3, GL_AMBIENT, light0_Ka);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, light0_Kd);
	glLightfv(GL_LIGHT3, GL_SPECULAR, light0_Ks);
	glLightfv(GL_LIGHT3, GL_POSITION, light3_Pos);
	glLightf(GL_LIGHT3, GL_CONSTANT_ATTENUATION, 0.0);
	glLightf(GL_LIGHT3, GL_LINEAR_ATTENUATION, 0.0);
	glLightf(GL_LIGHT3, GL_QUADRATIC_ATTENUATION, 0.0);

	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	// separate specular component
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
	glPopMatrix();

	lights_on[0] = true;
	lights_on[1] = false;
	lights_on[2] = false;
	lights_on[3] = false;

	if (enable_lighting)
	{
		DoEnableLights();
	}
	
	light_pos[0][0] = light0_Pos[0];
	light_pos[0][1] = light0_Pos[1];
	light_pos[0][2] = light0_Pos[2];
	light_pos[0][3] = light0_Pos[3];
	light_pos[1][0] = light1_Pos[0];
	light_pos[1][1] = light1_Pos[1];
	light_pos[1][2] = light1_Pos[2];
	light_pos[1][3] = light1_Pos[3];
	light_pos[2][0] = light2_Pos[0];
	light_pos[2][1] = light2_Pos[1];
	light_pos[2][2] = light2_Pos[2];
	light_pos[2][3] = light2_Pos[3];
	light_pos[3][0] = light3_Pos[0];
	light_pos[3][1] = light3_Pos[1];
	light_pos[3][2] = light3_Pos[2];
	light_pos[3][3] = light3_Pos[3];
	
	return;
}



void PDecompRenderer::InitMaterial(void)
{
	SetMaterial(MATERIAL_DEFAULT);
}

void PDecompRenderer::SetColorCodeType(int typeInd)
{
	this->colorCode_type = typeInd;	
}



void PDecompRenderer::SetMaterial(int material)
{
	GLfloat mat_Ka[4];
	GLfloat mat_Kd[4];
	GLfloat mat_Ks[4];
	GLfloat mat_Se;
	mat_Ka[3] = 1.0f;
	mat_Kd[3] = 1.0f;
	mat_Ks[3] = 1.0f;

	// Record material
	material_mode = material;

	switch(material)
	{
	case MATERIAL_BRASS:
		mat_Ka[0] = 0.329412f;
		mat_Kd[0] = 0.780392f;
		mat_Ks[0] = 0.992157f;
		mat_Ka[1] = 0.223529f;
		mat_Kd[1] = 0.568627f;
		mat_Ks[1] = 0.941176f;
		mat_Ka[2] = 0.027451f;
		mat_Kd[2] = 0.113725f;
		mat_Ks[2] = 0.807843f;
		mat_Se = 27.8974f;
		break;

	case MATERIAL_BRONZE:
		mat_Ka[0] = 0.2125f;
		mat_Kd[0] = 0.714f;
		mat_Ks[0] = 0.393548f;
		mat_Ka[1] = 0.1275f;
		mat_Kd[1] = 0.4284f;
		mat_Ks[1] = 0.271906f;
		mat_Ka[2] = 0.054f;
		mat_Kd[2] = 0.18144f;
		mat_Ks[2] = 0.166721f;
		mat_Se = 25.6f;
		break;

	case MATERIAL_CHROME:
		mat_Ka[0] = 0.25f;
		mat_Kd[0] = 0.4f;
		mat_Ks[0] = 0.774597f;
		mat_Ka[1] = 0.25f;
		mat_Kd[1] = 0.4f;
		mat_Ks[1] = 0.774597f;
		mat_Ka[2] = 0.25f;
		mat_Kd[2] = 0.4f;
		mat_Ks[2] = 0.774597f;
		mat_Se = 76.8f;
		break;

	case MATERIAL_COPPER:
		mat_Ka[0] = 0.19125f;
		mat_Kd[0] = 0.7038f;
		mat_Ks[0] = 0.256777f;
		mat_Ka[1] = 0.0735f;
		mat_Kd[1] = 0.27048f;
		mat_Ks[1] = 0.137622f;
		mat_Ka[2] = 0.0225f;
		mat_Kd[2] = 0.0828f;
		mat_Ks[2] = 0.086014f;
		mat_Se = 12.84f;
		break;

	case MATERIAL_GOLD:
		mat_Ka[0] = 0.24725f;
		mat_Kd[0] = 0.75164f;
		mat_Ks[0] = 0.628281f;
		mat_Ka[1] = 0.1995f;
		mat_Kd[1] = 0.60648f;
		mat_Ks[1] = 0.555802f;
		mat_Ka[2] = 0.0745f;
		mat_Kd[2] = 0.22648f;
		mat_Ks[2] = 0.366065f;
		mat_Se = 51.2f;
		break;

	case MATERIAL_P_GOLD:
		mat_Ka[0] = 0.24725f;
		mat_Kd[0] = 0.34615f;
		mat_Ks[0] = 0.797357f;
		mat_Ka[1] = 0.2245f;
		mat_Kd[1] = 0.3143f;
		mat_Ks[1] = 0.723991f;
		mat_Ka[2] = 0.0645f;
		mat_Kd[2] = 0.0903f;
		mat_Ks[2] = 0.208006f;
		mat_Se = 83.2f;
		break;

	case MATERIAL_PEWTER:
		mat_Ka[0] = 0.105882f;
		mat_Kd[0] = 0.427451f;
		mat_Ks[0] = 0.333333f;
		mat_Ka[1] = 0.058824f;
		mat_Kd[1] = 0.470588f;
		mat_Ks[1] = 0.333333f;
		mat_Ka[2] = 0.113725f;
		mat_Kd[2] = 0.541176f;
		mat_Ks[2] = 0.521569f;
		mat_Se = 9.84615f;
		break;

	case MATERIAL_SILVER:
		mat_Ka[0] = 0.19225f;
		mat_Kd[0] = 0.50754f;
		mat_Ks[0] = 0.508273f;
		mat_Ka[1] = 0.19225f;
		mat_Kd[1] = 0.50754f;
		mat_Ks[1] = 0.508273f;
		mat_Ka[2] = 0.19225f;
		mat_Kd[2] = 0.50754f;
		mat_Ks[2] = 0.508273f;
		mat_Se = 51.2f;
		break;

	case MATERIAL_P_SILVER:
		mat_Ka[0] = 0.23125f;
		mat_Kd[0] = 0.2775f;
		mat_Ks[0] = 0.773911f;
		mat_Ka[1] = 0.23125f;
		mat_Kd[1] = 0.2775f;
		mat_Ks[1] = 0.773911f;
		mat_Ka[2] = 0.23125f;
		mat_Kd[2] = 0.2775f;
		mat_Ks[2] = 0.773911f;
		mat_Se = 89.6f;
		break;

	case MATERIAL_EMERALD:
		mat_Ka[0] = 0.0215f;
		mat_Kd[0] = 0.07568f;
		mat_Ks[0] = 0.633f;
		mat_Ka[1] = 0.1745f;
		mat_Kd[1] = 0.61424f;
		mat_Ks[1] = 0.727811f;
		mat_Ka[2] = 0.0215f;
		mat_Kd[2] = 0.07568f;
		mat_Ks[2] = 0.633f;
		mat_Ka[3] = 0.55f;
		mat_Kd[3] = 0.55f;
		mat_Ks[3] = 0.55f;
		mat_Se = 76.8f;
		break;

	case MATERIAL_JADE:
		mat_Ka[0] = 0.135f;
		mat_Kd[0] = 0.54f;
		mat_Ks[0] = 0.316228f;
		mat_Ka[1] = 0.2225f;
		mat_Kd[1] = 0.89f;
		mat_Ks[1] = 0.316228f;
		mat_Ka[2] = 0.1575f;
		mat_Kd[2] = 0.63f;
		mat_Ks[2] = 0.316228f;
		mat_Ka[3] = 0.95f;
		mat_Kd[3] = 0.95f;
		mat_Ks[3] = 0.95f;
		mat_Se = 12.8f;
		break;

	case MATERIAL_OBSIDIAN:
		mat_Ka[0] = 0.05375f;
		mat_Kd[0] = 0.18275f;
		mat_Ks[0] = 0.332741f;
		mat_Ka[1] = 0.05f;
		mat_Kd[1] = 0.17f;
		mat_Ks[1] = 0.328634f;
		mat_Ka[2] = 0.06625f;
		mat_Kd[2] = 0.22525f;
		mat_Ks[2] = 0.346435f;
		mat_Ka[3] = 0.82f;
		mat_Kd[3] = 0.82f;
		mat_Ks[3] = 0.82f;
		mat_Se = 38.4f;
		break;

	case MATERIAL_PEARL:
		mat_Ka[0] = 0.25f;
		mat_Kd[0] = 1.0f;
		mat_Ks[0] = 0.296648f;
		mat_Ka[1] = 0.20725f;
		mat_Kd[1] = 0.829f;
		mat_Ks[1] = 0.296648f;
		mat_Ka[2] = 0.20725f;
		mat_Kd[2] = 0.829f;
		mat_Ks[2] = 0.296648f;
		mat_Ka[3] = 0.922f;
		mat_Kd[3] = 0.922f;
		mat_Ks[3] = 0.922f;
		mat_Se = 11.264f;
		break;

	case MATERIAL_RUDY:
		mat_Ka[0] = 0.1745f;
		mat_Kd[0] = 0.61424f;
		mat_Ks[0] = 0.727811f;
		mat_Ka[1] = 0.01175f;
		mat_Kd[1] = 0.04136f;
		mat_Ks[1] = 0.626959f;
		mat_Ka[2] = 0.01175f;
		mat_Kd[2] = 0.04136f;
		mat_Ks[2] = 0.626959f;
		mat_Ka[3] = 0.55f;
		mat_Kd[3] = 0.55f;
		mat_Ks[3] = 0.55f;
		mat_Se = 76.8f;
		break;

	case MATERIAL_TURQUOISE:
		mat_Ka[0] = 0.1f;
		mat_Kd[0] = 0.396f;
		mat_Ks[0] = 0.297254f;
		mat_Ka[1] = 0.18725f;
		mat_Kd[1] = 0.74151f;
		mat_Ks[1] = 0.30829f;
		mat_Ka[2] = 0.1745f;
		mat_Kd[2] = 0.69102f;
		mat_Ks[2] = 0.306678f;
		mat_Ka[3] = 0.8f;
		mat_Kd[3] = 0.8f;
		mat_Ks[3] = 0.8f;
		mat_Se = 12.8f;
		break;

	case MATERIAL_SKIN:
		mat_Ka[0] = 0.1f;
		mat_Ka[1] = 0.2f;
		mat_Ka[2] = 0.2f;
		mat_Kd[0] = 0.9f;
		mat_Kd[1] = 0.5f;
		mat_Kd[2] = 0.4f;
		mat_Ks[0] = 0.5f;
		mat_Ks[1] = 0.5f;
		mat_Ks[2] = 0.5f;
		mat_Se = 11.264f;
		break;


	case MATERIAL_INDIANRED:
		mat_Ka[0] = 0.2f;
		mat_Ka[1] = 0.2f;
		mat_Ka[2] = 0.2f;
		mat_Kd[0] = 1.0f;
		mat_Kd[1] = 0.42f;
		mat_Kd[2] = 0.42f;
		mat_Ks[0] = 0.5f;
		mat_Ks[1] = 0.5f;
		mat_Ks[2] = 0.5f;
		mat_Se = 15.0f;
		break;

	case MATERIAL_DODGEBLUE:
		mat_Ka[0] = 0.2f;
		mat_Ka[1] = 0.2f;
		mat_Ka[2] = 0.2f;
		mat_Kd[0] = 0.12f;
		mat_Kd[1] = 0.56f;
		mat_Kd[2] = 1.0f;
		mat_Ks[0] = 0.5f;
		mat_Ks[1] = 0.5f;
		mat_Ks[2] = 0.5f;
		mat_Se = 15.0f;
		break;

	case MATERIAL_DEFAULT:
	default:
		mat_Ka[0] = 0.3f;
		mat_Ka[1] = 0.3f;
		mat_Ka[2] = 0.3f;
		mat_Kd[0] = 1.0f;
		mat_Kd[1] = 1.0f;
		mat_Kd[2] = 1.0f;
		mat_Ks[0] = 0.7f;
		mat_Ks[1] = 0.7f;
		mat_Ks[2] = 0.7f;
		mat_Se = 64.0f;
		break;
		break;

	}

	glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_Kd);
	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_Ks);
	glMaterialfv(GL_FRONT, GL_AMBIENT, mat_Ka);
	glMaterialf(GL_FRONT, GL_SHININESS, mat_Se);

	mat_Ka[0] = 0.2f;
	mat_Ka[1] = 0.2f;
	mat_Ka[2] = 0.2f;
	mat_Kd[0] = 0.3f;  //backface color set here!!!
	mat_Kd[1] = 0.0f;
	mat_Kd[2] = 0.0f;
	mat_Ks[0] = 0.7f;
	mat_Ks[1] = 0.3f;
	mat_Ks[2] = 0.3f;
	mat_Se = 100.f;

	glMaterialfv(GL_BACK, GL_AMBIENT, mat_Ka);
	glMaterialfv(GL_BACK, GL_DIFFUSE, mat_Kd);
	glMaterialfv(GL_BACK, GL_SPECULAR, mat_Ks);
	glMaterialf(GL_BACK, GL_SHININESS, mat_Se);
}

void PDecompRenderer::SetLightsStatus(bool l0, bool l1, bool l2, bool l3)
{
	lights_on[0] = l0;
	lights_on[1] = l1;
	lights_on[2] = l2;
	lights_on[3] = l3;
	DoResetLights();
}

void PDecompRenderer::DoResetLights()
{
	DoDisableLights();
	DoEnableLights();
}

void PDecompRenderer::DoEnableLights()
{
	glEnable(GL_LIGHTING);

	if (lights_on[0])
		glEnable(GL_LIGHT0);

	if (lights_on[1])
		glEnable(GL_LIGHT1);

	if (lights_on[2])
		glEnable(GL_LIGHT2);

	if (lights_on[3])
		glEnable(GL_LIGHT3);
}

void PDecompRenderer::DoDisableLights()
{
	glDisable(GL_LIGHT0);

	glDisable(GL_LIGHT1);

	glDisable(GL_LIGHT2);

	glDisable(GL_LIGHT3);

	glDisable(GL_LIGHTING);
}



void PDecompRenderer::InitQuadrics()
{
	quadric_list = glGenLists(4);
	
	GLUquadricObj *qobj;
	qobj = gluNewQuadric();
	gluQuadricCallback(qobj, GLU_ERROR, (GluQuadricCallbackType)quadric_error_cb);

	gluQuadricDrawStyle(qobj, GLU_FILL);
	gluQuadricNormals(qobj, GLU_SMOOTH);

	// A sphere quadric
	glNewList(quadric_list, GL_COMPILE);
		gluSphere(qobj, 0.02, 30, 20);
	glEndList();

	// A cone quadric
	glNewList(quadric_list+1, GL_COMPILE);
		gluCylinder(qobj, 0.04, 0, 0.08, 30, 20);
	glEndList();

	// A disk quadric
	glNewList(quadric_list+2, GL_COMPILE);
		gluDisk(qobj, 0, 0.04, 30, 20);
	glEndList();
	
	// A cylinder quadric
	glNewList(quadric_list+3, GL_COMPILE);
		gluCylinder(qobj, 0.02, 0.02, 0.4, 30, 20);
	glEndList();
		
}


void PDecompRenderer::LoadTextures()
{
	std::cout << "Initiating textures...";
	InitTextureData();
	InitTexture2D();
	std::cout << "done" << std::endl;
}

void PDecompRenderer::SaveTextures()
{
	;
}

void PDecompRenderer::InitTextureData()
{
	m_ctexture2D.LoadFile("worldmap.bmp");
}

void PDecompRenderer::InitTexture2D()
{
	//glGenTextures(TEX2_COUNT, texture_name);

	//// Checkerboard Texture
	//glBindTexture(GL_TEXTURE_2D, texture_name[0]);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

	//glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
	//	0, GL_RGBA, GL_UNSIGNED_BYTE, check_texture2D);

	//if (gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
	//	GL_RGBA, GL_UNSIGNED_BYTE, check_texture2D) != 0)
	//	cerr << "Automatic 2D mipmap construction failed!" << endl;

	//// Environment mapping Texture
	//glBindTexture(GL_TEXTURE_2D, texture_name[1]);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

	//glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
	//	0, GL_RGBA, GL_UNSIGNED_BYTE, env_texture2D);

	//if (gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
	//	GL_RGBA, GL_UNSIGNED_BYTE, env_texture2D) != 0)
	//	cerr << "Automatic 2D mipmap construction failed!" << endl;

	//// Floor Texture
	//glBindTexture(GL_TEXTURE_2D, texture_name[2]);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

	//glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
	//	0, GL_RGBA, GL_UNSIGNED_BYTE, floor_texture2D);

	//if (gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
	//	GL_RGBA, GL_UNSIGNED_BYTE, floor_texture2D) != 0)
	//	cerr << "Automatic 2D mipmap construction failed!" << endl;
	//
	//glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	////
	glBindTexture(GL_TEXTURE_2D, m_ctexture2D.TexName());
}


void PDecompRenderer::InitCamera()
{
	vec3 homepos, homerotaxis;

	// Since the model is always normalized, constants suffice
	homepos.x = 0;
	homepos.y = 0;
	homepos.z = 3.0;
	
	homerotaxis = vec3(0, 0, 1);

	theCamera.SetHome(homepos, 0, homerotaxis, DEFAULT_FOV);
	rotate_dist = view_dist = 3.0;

	theCamera.GoHome();
}



void PDecompRenderer::InitTraits(int mInd)
{
	if (!mesh[mInd]->hasNormal())
	{
		GetNormals::GetNorm(mesh[cMeshInd], true);
	}		
}


// Right now, the models are always normalized and the camera setting is fixed
// TODO: instead of normalizing the models, it may be more appropriate to adjust 
// the camera setting automatically for different models
void PDecompRenderer::SetModel(Mesh *pMesh, int mInd)
{
	assert(mInd >= 0);
	if (!pMesh)
		return;
	if (mInd >= mesh.size())
		mesh.resize(mInd, NULL);
	mesh[mInd]  = pMesh;
	MeshUtility::MarkSharpEdges(mesh[mInd]);
	InitTraits(mInd);
}


// Change the model to be displayed
// Camera setting is unchanged
//void PDecompRenderer::ChangeModel(Mesh *pMesh)
//{
//	if (pMesh != NULL)
//		mesh[cMeshInd] = pMesh;
//	InitTraits(cMeshInd);
//}

void PDecompRenderer::ComputeBoundingBox(int meshInd)
{
	assert(meshInd >= 0 && meshInd <= 3);
	Mesh * cmesh = mesh[meshInd];

	if (!cmesh)
		return;

	double minx, maxx, miny, maxy, minz, maxz;
	minx = miny = minz = 1e10;
	maxx = maxy = maxz = -1e10;

	for (MeshVertexIterator viter(cmesh); !viter.end(); ++viter)
	{
		Vertex *v = *viter;
		if (v->point()[0] < minx) minx = v->point()[0];
		if (v->point()[0] > maxx) maxx = v->point()[0];

		if (v->point()[1] < miny) miny = v->point()[1];
		if (v->point()[1] > maxy) maxy = v->point()[1];

		if (v->point()[2] < minz) minz = v->point()[2];
		if (v->point()[2] > maxz) maxz = v->point()[2];
	}

	if (m_bbox.size() <= meshInd)
	{
		for (int i=m_bbox.size(); i<=meshInd; ++i)
		{
			Boundingboxd dummy;
			m_bbox.push_back(dummy);
		}			
	}

	m_bbox[meshInd].minX() = minx; m_bbox[meshInd].maxX() = maxx;
	m_bbox[meshInd].minY() = miny; m_bbox[meshInd].maxY() = maxy;
	m_bbox[meshInd].minZ() = minz, m_bbox[meshInd].maxZ() = maxz;

}

void PDecompRenderer::SetBoundingBox(const Boundingboxd & bbox, int box_id)
{
	assert(box_id >= 0 && box_id <= 3);
	m_bbox[box_id] = bbox;
}


// The difference between 'RenderGL' and 'draw' is that if double-buffering is enabled in the window, 
// the back and front buffers are swapped after 'draw' is completed
void PDecompRenderer::RenderGL()
{
	assert(initialized);
	Reshape(w(), h());
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	theCamera.SetGL(view_dist/DOF, view_dist*DOF, win_width, win_height);

	if (shading_mode == SHADING_SMOOTH)
		glShadeModel(GL_SMOOTH);
	else 
		glShadeModel(GL_FLAT);

	//if (show_psetrect)
	//{
	//	RenderPantsSetRect(m_pantsSet);
	//}
	if (show_morph)
	{
		RenderMorphing(m_pantsSet);
	}
	else if (show_hexagons)
	{
		RenderHexagons(m_pantsSet);
	}
	else if (show_pset)
	{
		RenderPantsSet(m_pantsSet);
	}
	else if (show_sphere)
	{
		RenderSphere(*pd);
	}
	else
	{
		renderingMesh = mesh[cMeshInd];
		RenderMesh(renderingMesh);
		if (show_sharpvert)
			RenderSharpVertex(renderingMesh);		
		if (show_homo && pd && pd->Status()>=0)
			RenderHomoBasis(*pd);
		if (!cMeshInd && pd)
			RenderGuideCurve();
		if (show_sharpedge)
			RenderSharpEdge(renderingMesh);			
		if (show_triangulation)
			RenderTessellation(renderingMesh);
		if (show_normal)
			RenderNormal(renderingMesh);
		if (show_boundingbox)
			RenderBoundingBox(cMeshInd);
		if (show_floor)
			RenderFloor(cMeshInd);
	}


	glPopMatrix();
}


void PDecompRenderer::RenderMesh(Mesh *pMesh, int useRGB)
{
	if (!pMesh) 
		return;

	int cur_mat = material_mode;

	float crgb[3];

	if (useRGB == -1)
	{
		glDisable(GL_COLOR_MATERIAL);
		crgb[0] = surfaceColor[0];
		crgb[1] = surfaceColor[1];
		crgb[2] = surfaceColor[2];
	}
	else
	{
		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT, GL_DIFFUSE);		
		RGBBar::GetRGB_Float(useRGB, crgb);		
	}

	SetMaterial(cur_mat);

	if (show_envmap)
	{
		BeginEnvMap();
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texture2_name[1]);
	}
	else if (show_texcoord)
	{		
		glDisable(GL_COLOR_MATERIAL);
		SetMaterial(cur_mat);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, m_ctexture2D.TexName());//texture_name[0]);
	}

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glPolygonMode(GL_FRONT, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_LINE:
		glPolygonMode(GL_FRONT, GL_LINE);
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_POINT:
		glPolygonMode(GL_FRONT, GL_POINT);
		glEnable(GL_POLYGON_OFFSET_POINT);
		glPolygonOffset(1.0, 1.0);
		glPointSize(3.0);
		break;
	}

	if (!pMesh->hasNormal())
	{
		GetNormals::GetNorm(pMesh, true);
	}

	bool hasTUV = pMesh->hasUV();
	if (show_texcoord && !pMesh->hasUV())
	{
		hasTUV = GetUV::ParseUV(pMesh);
	}

	glColor3f(crgb[0], crgb[1], crgb[2]);
	
	glBegin(GL_TRIANGLES);

	for(MeshFaceIterator fiter(pMesh); !fiter.end(); ++fiter)
	{
		Face *f = *fiter;

		// sharp face
		if (show_sharpface && f->sharp())
		{
			DoDisableLights();
			glColor3f(sharpFaceColor[0], sharpFaceColor[1], sharpFaceColor[2]);
		}

		Vertex * v[3];
		HalfEdge * he = f->halfedge();
		v[0] = he->source();
		v[1] = he->target();
		v[2] = he->he_next()->target();

		Point * p[3];
		Point * n[3];

		p[0] = &v[0]->point();
		p[1] = &v[1]->point();
		p[2] = &v[2]->point();

		n[0] = &pMesh->verNormal(v[0]);
		n[1] = &pMesh->verNormal(v[1]);
		n[2] = &pMesh->verNormal(v[2]);

		double uv[3][3];
	
		if (show_texcoord && hasTUV)
		{
			Point & puv0 = pMesh->verUV(v[0]);
			uv[0][0] = puv0[0]; uv[0][1] = puv0[1]; uv[0][2] = puv0[2];
			Point & puv1 = pMesh->verUV(v[1]);
			uv[1][0] = puv1[0]; uv[1][1] = puv1[1]; uv[1][2] = puv1[2];
			Point & puv2 = pMesh->verUV(v[2]);
			uv[2][0] = puv2[0]; uv[2][1] = puv2[1]; uv[2][2] = puv2[2];
			UV2TexUV(uv[0], uv[1], uv[2], uv[0], uv[1], uv[2]);
		}
		
		for (int i=0;i<3;++i)
		{
			if (show_texcoord && hasTUV) // && textype)
				glTexCoord2d(uv[i][0], uv[i][1]);
			glNormal3d( (*n[i])[0], (*n[i])[1], (*n[i])[2] );
			glVertex3d( (*p[i])[0], (*p[i])[1], (*p[i])[2] );
		}		

		if (show_sharpface && f->sharp())
		{
			if (enable_lighting)
			{
				DoEnableLights();
			}
			glColor3f(crgb[0], crgb[1], crgb[2]);
		}

	}
	
	glEnd();

	glPolygonMode(GL_FRONT, GL_FILL);

	if (glIsEnabled(GL_TEXTURE_2D))
		glDisable(GL_TEXTURE_2D);

	if (show_envmap)
		EndEnvMap();

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glDisable(GL_POLYGON_OFFSET_FILL);
		break;
	case POLYGON_LINE:
		glDisable(GL_POLYGON_OFFSET_LINE);
		break;
	case POLYGON_POINT:
		glDisable(GL_POLYGON_OFFSET_POINT);
		glPointSize(1.0);
		break;
	}

}

void PDecompRenderer::RenderPantsSet(PantsSet * pSet)
{
	if (show_psetseparate)
	{//render one
		renderingMesh = pSet->IDPant(show_pantInd)->PantMesh(); 
		RenderMesh(renderingMesh, show_pantInd);
		if (show_sharpvert)
			RenderSharpVertex(renderingMesh);		
		if (show_sharpedge)
			RenderSharpEdge(renderingMesh);
		if (show_triangulation)
			RenderTessellation(renderingMesh);
		if (show_normal)
			RenderNormal(renderingMesh);
		if (show_psetboundary)		
			RenderPantsSetBoundary(m_pantsSet, show_pantInd);
		if (show_psetbmarker)
			RenderPantsSetBMarker(m_pantsSet, show_pantInd);
		else if (show_pantsfeatures)
			RenderPantsSetFeatures(m_pantsSet, show_pantInd);
	}
	else
	{//render all
		for (int i=0;i<pSet->PantSize(); ++i)
		{
			renderingMesh = pSet->IDPant(i)->PantMesh(); 
			RenderMesh(renderingMesh, i);
			if (show_sharpvert)
				RenderSharpVertex(renderingMesh);		
			if (show_sharpedge)
				RenderSharpEdge(renderingMesh);
			if (show_triangulation)
				RenderTessellation(renderingMesh);
			if (show_normal)
				RenderNormal(renderingMesh);
			if (show_psetboundary)		
				RenderPantsSetBoundary(m_pantsSet, i);	
			if (show_psetbmarker)
				RenderPantsSetBMarker(m_pantsSet, i);
			else if (show_pantsfeatures)
			{
				if (i < 6)
					RenderPantsSetFeatures(m_pantsSet, i);
			}			
		}
	}
}

void PDecompRenderer::RenderHexagon(Pant *cPant)
{
	if (!show_hexagonUV)
	{
		for (int i=0;i<2;++i)
		{
			renderingMesh = cPant->GetPatch(i);
			RenderMesh(renderingMesh, cPant->ID() + i*m_pantsSet->PantSize());
			if (show_sharpvert)
				RenderSharpVertex(renderingMesh);		
			if (show_sharpedge)
				RenderSharpEdge(renderingMesh);
			if (show_triangulation)
				RenderTessellation(renderingMesh);
			if (show_normal)
				RenderNormal(renderingMesh);
			if (show_psetboundary)		
				RenderPantsSetBoundary(m_pantsSet, cPant->ID());
			if (show_psetbmarker)
				RenderPantsSetBMarker(m_pantsSet, cPant->ID());
		}
	}
	else
	{
		//
	}
}

void PDecompRenderer::RenderHexagonCFKM(Pant *cPant)
{
	std::string strpos = "rmpos";
	std::string strnorm = "rmnormal";
	double value[3];
	if (!show_hexagonUV)
	{
		for (int i=0;i<2;++i)
		{
			if (!psMap)
				return;
			stdext::hash_map<Vertex *, Point> * rgbList;
			if (show_KMRGB)
				rgbList = &(psMap->kmRGBLists[i][cPant->ID()]);
			else 
			{
				assert(show_CFRGB);
				rgbList = &(psMap->cfRGBLists[i][cPant->ID()]);
			}

			renderingMesh = cPant->GetPatch(i);
			
			//render the mesh here!!

			if (!renderingMesh) 
				return;

			int cur_mat = material_mode;

			//glEnable(GL_COLOR_MATERIAL);
			//glColorMaterial(GL_FRONT, GL_DIFFUSE);

			glEnable(GL_COLOR_MATERIAL);
			glColorMaterial(GL_FRONT, GL_DIFFUSE);		

			SetMaterial(cur_mat);

			switch(polygon_mode)
			{
			case POLYGON_FILL:
				glPolygonMode(GL_FRONT, GL_FILL);
				glEnable(GL_POLYGON_OFFSET_FILL);
				glPolygonOffset(1.0, 1.0);
				break;
			case POLYGON_LINE:
				glPolygonMode(GL_FRONT, GL_LINE);
				glEnable(GL_POLYGON_OFFSET_LINE);
				glPolygonOffset(1.0, 1.0);
				break;
			case POLYGON_POINT:
				glPolygonMode(GL_FRONT, GL_POINT);
				glEnable(GL_POLYGON_OFFSET_POINT);
				glPolygonOffset(1.0, 1.0);
				glPointSize(3.0);
				break;
			}

			if (!renderingMesh->hasNormal())
			{
				GetNormals::GetNorm(renderingMesh, true);
			}
			bool useTrgPos = true;
			std::vector<Point> TrgPosArray;
			std::vector<Point> TrgNormArray;
			if (useTrgPos)
			{
				for (MeshVertexIterator viter(renderingMesh); !viter.end(); ++viter)
				{
					Vertex * cver = *viter;
					//assert(renderingMesh->id2VInd[cver->id()] == TrgPosArray.size());					
					TraitParser::ReadTrait(cver->string(), strpos, 3, value);
					TrgPosArray.push_back(Point(value[0],value[1],value[2]));
					TraitParser::ReadTrait(cver->string(), strnorm, 3, value);
					TrgNormArray.push_back(Point(value[0],value[1],value[2]));
				}
			}

			glBegin(GL_TRIANGLES);

			for(MeshFaceIterator fiter(renderingMesh); !fiter.end(); ++fiter)
			{
				Face *f = *fiter;

				Vertex * v[3];
				HalfEdge * he = f->halfedge();
				v[0] = he->source();
				v[1] = he->target();
				v[2] = he->he_next()->target();

				Point * p[3];
				Point * n[3];				
				
				if (!useTrgPos)
				{
					p[0] = &(v[0]->point());
					p[1] = &(v[1]->point());
					p[2] = &(v[2]->point());
					n[0] = &(renderingMesh->verNormal(v[0]));
					n[1] = &(renderingMesh->verNormal(v[1]));
					n[2] = &(renderingMesh->verNormal(v[2]));
				}
				else
				{
					int cvind0 = renderingMesh->id2VInd[v[0]->id()];
					int cvind1 = renderingMesh->id2VInd[v[1]->id()];
					int cvind2 = renderingMesh->id2VInd[v[2]->id()];
					p[0] = &(TrgPosArray[cvind0]);
					p[1] = &(TrgPosArray[cvind1]);
					p[2] = &(TrgPosArray[cvind2]);
					n[0] = &(TrgNormArray[cvind0]);
					n[1] = &(TrgNormArray[cvind1]);
					n[2] = &(TrgNormArray[cvind2]);					
				}				
					
				for (int i=0;i<3;++i)
				{
					Point & crgb = (*rgbList)[v[i]];
					glColor3f(crgb[0], crgb[1], crgb[2]);
					glNormal3d( (*n[i])[0], (*n[i])[1], (*n[i])[2] );
					glVertex3d( (*p[i])[0], (*p[i])[1], (*p[i])[2] );
				}	

			}
		
			glEnd();

			//rendering boundaries
			if (show_psetboundary)
			{
				DoDisableLights();
			
				// anti-aliasing
				glEnable (GL_LINE_SMOOTH);
				glEnable (GL_BLEND);
				glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

				glLineWidth(2.0);
				glColor3f(sharpEdgeColor[0], sharpEdgeColor[1], sharpEdgeColor[2]);

				glBegin(GL_LINES);		
				for(MeshEdgeIterator eiter(renderingMesh); !eiter.end(); ++eiter)
				{
					Edge * e = *eiter;
					if (!e->sharp() && !e->boundary())
						continue;

					Vertex * ev1 = renderingMesh->edgeVertex1(e);
					Vertex * ev2 = renderingMesh->edgeVertex2(e);					

					Point & ep1 = ev1->point(); 
					Point & ep2 = ev2->point();

					glVertex3f(ep1[0], ep1[1], ep1[2]);
					glVertex3f(ep2[0], ep2[1], ep2[2]);	
				}
				glEnd();
				glLineWidth(1.0);

				if (enable_lighting)
				{
					DoEnableLights();
				}

				glDisable (GL_LINE_SMOOTH);
				glDisable (GL_BLEND);

			}

			glPolygonMode(GL_FRONT, GL_FILL);

			switch(polygon_mode)
			{
			case POLYGON_FILL:
				glDisable(GL_POLYGON_OFFSET_FILL);
				break;
			case POLYGON_LINE:
				glDisable(GL_POLYGON_OFFSET_LINE);
				break;
			case POLYGON_POINT:
				glDisable(GL_POLYGON_OFFSET_POINT);
				glPointSize(1.0);
				break;
			}


			//RenderMesh(renderingMesh, cPant->ID() + i*m_pantsSet->PantSize());
			//if (show_sharpvert)
			//	RenderSharpVertex(renderingMesh);		
			//if (show_sharpedge)
			//	RenderSharpEdge(renderingMesh);
			//if (show_triangulation)
			//	RenderTessellation(renderingMesh);
			//if (show_normal)
			//	RenderNormal(renderingMesh);
			//if (show_psetboundary)		
			//	RenderPantsSetBoundary(m_pantsSet, cPant->ID());
			//if (show_psetbmarker)
			//	RenderPantsSetBMarker(m_pantsSet, cPant->ID());
		}
	}
	else
	{
		//
	}
}

void PDecompRenderer::RenderMorphing(Pant * cPant, const int useRGB)
{

	int cur_mat = material_mode;
	
	float crgb[3];

	if (useRGB == -1)
	{
		glDisable(GL_COLOR_MATERIAL);
		crgb[0] = surfaceColor[0];
		crgb[1] = surfaceColor[1];
		crgb[2] = surfaceColor[2];
	}
	else
	{
		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT, GL_DIFFUSE);		
		RGBBar::GetRGB_Float(useRGB, crgb);		
	}

	glColor3f(crgb[0], crgb[1], crgb[2]);


	SetMaterial(cur_mat);

	if (show_envmap)
	{
		BeginEnvMap();
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texture2_name[1]);
	}
	else if (show_texcoord)
	{		
		glDisable(GL_COLOR_MATERIAL);
		SetMaterial(cur_mat);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, m_ctexture2D.TexName());//texture_name[0]);
	}

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glPolygonMode(GL_FRONT, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_LINE:
		glPolygonMode(GL_FRONT, GL_LINE);
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_POINT:
		glPolygonMode(GL_FRONT, GL_POINT);
		glEnable(GL_POLYGON_OFFSET_POINT);
		glPolygonOffset(1.0, 1.0);
		glPointSize(3.0);
		break;
	}

	glBegin(GL_TRIANGLES);

	for (int i=0;i<2;++i)
	{
		Mesh * pMesh = cPant->GetPatch(i);
		if (!pMesh) 
			return;

		if (!pMesh->hasNormal())
		{
			GetNormals::GetNorm(pMesh, true);
		}

		bool hasTUV = pMesh->hasUV();
		if (show_texcoord && !pMesh->hasUV())
		{
			hasTUV = GetUV::ParseUV(pMesh);
		}		

		for(MeshFaceIterator fiter(pMesh); !fiter.end(); ++fiter)
		{
			Face *f = *fiter;

			// sharp face
			if (show_sharpface && f->sharp())
			{
				DoDisableLights();
				glColor3f(sharpFaceColor[0], sharpFaceColor[1], sharpFaceColor[2]);
			}

			Vertex * v[3];
			HalfEdge * he = f->halfedge();
			v[0] = he->source();
			v[1] = he->target();
			v[2] = he->he_next()->target();
			PantVertexTrait *pvt[3];
			pvt[0] = (PantVertexTrait *)v[0]->GetTrait(ID_PantVertexTrait);
			pvt[1] = (PantVertexTrait *)v[1]->GetTrait(ID_PantVertexTrait);
			pvt[2] = (PantVertexTrait *)v[2]->GetTrait(ID_PantVertexTrait);
			if (!pvt[0] || !pvt[1] || !pvt[2])
			{
				std::cerr << "Error: pant vertex traits are not provided!" << std::endl;
				return ;
			}			

			Point * p[3];
			Point * n[3];

			p[0] = &v[0]->point();
			p[1] = &v[1]->point();
			p[2] = &v[2]->point();

			n[0] = &pMesh->verNormal(v[0]);
			n[1] = &pMesh->verNormal(v[1]);
			n[2] = &pMesh->verNormal(v[2]);

			Point * trgP[3];
			Point * trgN[3];

			trgP[0] = &pvt[0]->RemeshPos();
			trgP[1] = &pvt[1]->RemeshPos();
			trgP[2] = &pvt[2]->RemeshPos();

			trgN[0] = &pvt[0]->RemeshNormal();
			trgN[1] = &pvt[1]->RemeshNormal();
			trgN[2] = &pvt[2]->RemeshNormal();

			double uv[3][3];
		
			if (show_texcoord && hasTUV)
			{
				Point & puv0 = pMesh->verUV(v[0]);
				uv[0][0] = puv0[0]; uv[0][1] = puv0[1]; uv[0][2] = puv0[2];
				Point & puv1 = pMesh->verUV(v[1]);
				uv[1][0] = puv1[0]; uv[1][1] = puv1[1]; uv[1][2] = puv1[2];
				Point & puv2 = pMesh->verUV(v[2]);
				uv[2][0] = puv2[0]; uv[2][1] = puv2[1]; uv[2][2] = puv2[2];
				UV2TexUV(uv[0], uv[1], uv[2], uv[0], uv[1], uv[2]);
			}

			for (int i=0;i<3;++i)
			{
				if (show_texcoord && hasTUV) // && textype)
					glTexCoord2d(uv[i][0], uv[i][1]);

				Point cPos = *(p[i]) * (1-morphing_coef) + *(trgP[i]) * morphing_coef;
				Point cNorm = *(n[i]) *(1-morphing_coef) + *(trgN[i]) * morphing_coef;

				glNormal3d( cNorm[0], cNorm[1], cNorm[2] );
				glVertex3d( cPos[0], cPos[1], cPos[2] );
			}


			if (show_sharpface && f->sharp())
			{
				if (enable_lighting)
				{
					DoEnableLights();
				}
				glColor3f(crgb[0], crgb[1], crgb[2]);
			}

		}
	}
	
	glEnd();

	glPolygonMode(GL_FRONT, GL_FILL);

	if (glIsEnabled(GL_TEXTURE_2D))
		glDisable(GL_TEXTURE_2D);

	if (show_envmap)
		EndEnvMap();

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glDisable(GL_POLYGON_OFFSET_FILL);
		break;
	case POLYGON_LINE:
		glDisable(GL_POLYGON_OFFSET_LINE);
		break;
	case POLYGON_POINT:
		glDisable(GL_POLYGON_OFFSET_POINT);
		glPointSize(1.0);
		break;
	}


}

void PDecompRenderer::RenderMorphing(PantsSet * pSet)
{
	if (!pSet)
		return;
	if (show_psetseparate)
	{//render one
		Pant * cpant = pSet->IDPant(show_pantInd);
		RenderMorphing(cpant, show_pantInd);
	}
	else
	{
		for (int i=0;i<pSet->PantSize();++i)
		{
			Pant * cpant = pSet->IDPant(i);
			RenderMorphing(cpant, i);
		}
	}
}

void PDecompRenderer::RenderHexagons(PantsSet * pSet)
{
	if (show_psetseparate)
	{//render one
		Pant * cpant = pSet->IDPant(show_pantInd);
		if (show_CFRGB || show_KMRGB)
			RenderHexagonCFKM(cpant);
		else
			RenderHexagon(cpant);
	}
	else
	{//render all
		for (int i=0;i<pSet->PantSize(); ++i)
		{
			Pant * cpant = pSet->IDPant(i);
			if (show_CFRGB || show_KMRGB)
				RenderHexagonCFKM(cpant);
			else
				RenderHexagon(cpant);
		}
	}
}

void PDecompRenderer::RenderPantsSetBoundary(PantsSet * pSet, const int pantInd)
{
	Pant * cpant = pSet->IDPant(pantInd);
	Curve * nc[3];

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);	

	// Draw the hidden lines (edges)
	DoDisableLights();

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glLineWidth(2.0);	

	glBegin(GL_LINES);
	for (int i=0;i<3;++i)
	{
		float crgb[3];
		crgb[0] = crgb[1] = crgb[2] = 0;
		crgb[i] = 1;
		glColor3f(crgb[0], crgb[1], crgb[2]);
		nc[i] = cpant->PantBoundary(i);
		for (int j=0;j<nc[i]->heSize(); ++j)
		{
			HalfEdge * he = (*nc[i])[j];
			Point & p1 = he->source()->point();
			Point & p2 = he->target()->point();
			glVertex3f(p1[0], p1[1], p1[2]);
			glVertex3f(p2[0], p2[1], p2[2]);
		}
	}
	
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glDisable (GL_LINE_SMOOTH);
	glDisable (GL_BLEND);

	glDisable(GL_COLOR_MATERIAL);
}

void PDecompRenderer::RenderPantsSetFeatures(PantsSet * pSet, const int pantInd)
{
	Pant * cpant = pSet->IDPant(pantInd);

	if (!cpant)
		return;

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	DoDisableLights();

	float radius = m_bbox[0].diag() / 200 * 50;


	float maxRadius = 1.0f;
	float minRadius = 0.8f;

		
	if (radius > maxRadius) radius = maxRadius;
	if (radius < minRadius) radius = minRadius;

	float crgb[6][3] = {
		1,0,0,
		1,0,0,
		1,0,0,
		0,1,1,
		0,1,1,
		0,1,1};
			
	for (int i=0;i<6;++i)
	{
		if (i != 2 && i!= 4)
			continue;
		Vertex * ver = cpant->Marker(i);
		if (ver)
		{
			Point & p = ver->point();;
			glPushMatrix();
				glTranslated(p[0], p[1], p[2]);
				RenderBall(crgb[i][0], crgb[i][1], crgb[i][2], radius);
			glPopMatrix();
		}
	}
	
	if (enable_lighting)
	{
		DoEnableLights();
	}
	glDisable(GL_COLOR_MATERIAL);

}

void PDecompRenderer::RenderPantsSetBMarker(PantsSet * pSet, const int pantInd)
{
	Pant * cpant = pSet->IDPant(pantInd);

	if (!cpant)
		return;

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	DoDisableLights();

	float radius = m_bbox[0].diag() / 200 * 50;

	if (radius > 0.5f) radius = 0.5f;
	if (radius < 0.2f) radius = 0.2f;

	float crgb[6][3] = {
		1,0,0,
		0,1,0,
		0,0,1,
		1,1,0,
		1,0,1,
		0,1,1};
			
	for (int i=0;i<6;++i)
	{
		Vertex * ver = cpant->Marker(i);
		if (ver)
		{
			Point & p = ver->point();;
			glPushMatrix();
				glTranslated(p[0], p[1], p[2]);
				RenderBall(crgb[i][0], crgb[i][1], crgb[i][2], radius);
			glPopMatrix();
		}
	}
	
	if (enable_lighting)
	{
		DoEnableLights();
	}
	glDisable(GL_COLOR_MATERIAL);

}

//void PDecompRenderer::RenderPantsSetRect(PantsSet * pSet)
//{
//	for (int i=0;i<pSet->PantSize(); ++i)
//	{
//		for (int j=0;j<4;++j)
//		{
//			Mesh * cmesh = pSet->IDPant(i)->GetPatch(j); 
//			RenderMesh(cmesh, i*4+j);
//			if (show_sharpvert)
//				RenderSharpVertex(cmesh);		
//			if (show_sharpedge)
//				RenderSharpEdge(cmesh);
//			if (show_triangulation)
//				RenderTessellation(cmesh);
//		}
//	}
//}

void PDecompRenderer::RenderTessellation(Mesh * pMesh)
{
}

void PDecompRenderer::RenderSharpVertex(Mesh *pMesh)
{
	if (!pMesh)
		return;

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	DoDisableLights();

	float radius = m_bbox[0].diag() / 200 * 50;

	if (radius > 0.7f) radius = 0.7f;
	if (radius < 0.3f) radius = 0.3f;

	if (selectedVertices.size())
	{//faster way
		for (int i=0;i<selectedVertices.size();++i)
		{
			Vertex * ver = selectedVertices[i];
			Point & p = ver->point();
			//float crgb[3];			
			//RGBBar::GetRGB_Float(i, crgb);
			glPushMatrix();
				glTranslated(p[0], p[1], p[2]);
				RenderBall(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2], radius);				
				//RenderBall(crgb[0], crgb[1], crgb[2], radius);
			glPopMatrix();
			//if (show_markers)
			//{
			//	char text[10];
			//	sprintf(text, "%d", i);
			//	RenderTextOnPoint(text, p);
			//}
		}
	}
	else
	{//slow way

		FeaturePoints & cfps = pd->FeaturePts();
		for (int i=0;i<cfps.size();++i)
		{
			Vertex * ver = cfps.FeaturePoint(i);
			if (renderingMesh->hasVertex(ver))
			{
				Point & p = ver->point();;
				//float crgb[3];
				//RGBBar::GetRGB_Float(i, crgb);
				glPushMatrix();
					glTranslated(p[0], p[1], p[2]);
					RenderBall(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2], radius);
					//RenderBall(crgb[0], crgb[1], crgb[2], radius);
				glPopMatrix();	
			}
		}

		std::vector<SurgeryPoints *> & csplist = pd->SurgeryPtsList();
		for (int i=0;i<csplist.size();++i)
		{
			SurgeryPoints * csp = csplist[i];
			if (csp)
			{
				for (int j=0;j<csp->size();++j)
				{
					Vertex * ver = csp->FeaturePoint(j);
					int cHandleInd = csp->HandleInd();
					if (renderingMesh->hasVertex(ver))
					{
						Point & p = ver->point();;
						//float crgb[3];
						//RGBBar::GetRGB_Float(i, crgb);
						glPushMatrix();
							glTranslated(p[0], p[1], p[2]);
							RenderBall(HomoRGB[cHandleInd][0], HomoRGB[cHandleInd][1], HomoRGB[cHandleInd][2], radius);
							//RenderBall(crgb[0], crgb[1], crgb[2], radius);
						glPopMatrix();	
					}
				}
			}
		}

		

		//std::string fstr = "feature";
		//int value[1];
		//for (MeshVertexIterator viter(pMesh); !viter.end(); ++viter)
		//{
		//	Vertex * ver= *viter;
		//	bool isFeature = TraitParser::ReadIntTrait(ver->string(), fstr, 1, value);
		//	if (isFeature)
		//	{
		//		Point & p = ver->point();
		//		float crgb[3];
		//		RGBBar::GetRGB_Float(value[0], crgb);
		//		glPushMatrix();
		//			glTranslated(p[0], p[1], p[2]);
		//			RenderBall(crgb[0], crgb[1], crgb[2], radius);
		//		glPopMatrix();	
		//	}
		//}
	}


	//
	if (enable_lighting)
	{
		DoEnableLights();
	}
	glDisable(GL_COLOR_MATERIAL);


}

void PDecompRenderer::RenderSharpEdge(Mesh *pMesh)
{
	if (!pMesh) 
		return;

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);	

	// Draw the hidden lines (edges)
	DoDisableLights();

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glLineWidth(2.0);
	glColor3f(sharpEdgeColor[0], sharpEdgeColor[1], sharpEdgeColor[2]);

	glBegin(GL_LINES);		
	for(MeshEdgeIterator eiter(pMesh); !eiter.end(); ++eiter)
	{
		Edge * e = *eiter;
		if (!e->sharp())
			continue;

		Vertex * v1 = pMesh->edgeVertex1(e);
		Vertex * v2 = pMesh->edgeVertex2(e);
		Point p1, p2;

		p1 = v1->point(); 
		p2 = v2->point();

		glVertex3f(p1[0], p1[1], p1[2]);
		glVertex3f(p2[0], p2[1], p2[2]);	
	}
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glDisable (GL_LINE_SMOOTH);
	glDisable (GL_BLEND);

	glDisable(GL_COLOR_MATERIAL);
}

void PDecompRenderer::RenderHomoBasis(PantsDecomposition &pd)
{
	if (pd.Status()!=0 && pd.Status()!=1)
		return;
	//assure its status is 0

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);	

	// Draw the hidden lines (edges)
	DoDisableLights();

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glLineWidth(3.0);
	
	glBegin(GL_LINES);
	if (pd.Status()==0)
	{//render homo
		assert(pd.Genus() <= 9);
		for (int i=0;i<pd.Genus();++i)
		{		
			glColor3f(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2]);
			Curve * hl = pd.HandleLoop(i);
			for (int j=0;j<hl->heSize();++j)
			{
				HalfEdge * he = (*hl)[j];
				Point & p1 = he->source()->point();
				Point & p2 = he->target()->point();
				glVertex3f(p1[0], p1[1], p1[2]);
				glVertex3f(p2[0], p2[1], p2[2]);	
			}
			Curve * tl = pd.TunnelLoop(i);
			for (int j=0;j<tl->heSize();++j)
			{
				HalfEdge * he = (*tl)[j];
				Point & p1 = he->source()->point();
				Point & p2 = he->target()->point();
				glVertex3f(p1[0], p1[1], p1[2]);
				glVertex3f(p2[0], p2[1], p2[2]);	
			}
		}
	}
	else
	{//render homobound
		assert(pd.VGenus() <= 9);
		for (int i=0;i<pd.VGenus();++i)
		{		
			glColor3f(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2]);
			Curve * homoC = pd.HomoBoundary(i);
			for (int j=0;j<homoC->heSize();++j)
			{
				HalfEdge * he = (*homoC)[j];
				Point & p1 = he->source()->point();
				Point & p2 = he->target()->point();
				glVertex3f(p1[0], p1[1], p1[2]);
				glVertex3f(p2[0], p2[1], p2[2]);	
			}
		}
	}
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glDisable (GL_LINE_SMOOTH);
	glDisable (GL_BLEND);

	glDisable(GL_COLOR_MATERIAL);
}

void PDecompRenderer::RenderSphere(PantsDecomposition &pd)
{
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);

	//render vectors

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glColorMaterial(GL_FRONT, GL_SPECULAR);
	glColor3f(0.7f, 0.7f, 0.7f);
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(0.3f, 0.3f, 0.3f);

	assert(pd.Genus() <= 9);
	glLineWidth(4.0);
	DoDisableLights();
	glBegin(GL_LINES);
	for (int i=0;i<pd.TunnelMass().size(); ++i)
	{
		Point & pt = pd.TunnelMass(i);
		glColor3f(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2]);
		glVertex3d(pd.Mass()[0], pd.Mass()[1], pd.Mass()[2]);
		glVertex3d(pt[0], pt[1], pt[2]);
	}
	glEnd();
	glLineWidth(1.0);
	if (enable_lighting)
	{
		DoEnableLights();
	}		

	//render balls
	glPushMatrix();
	glTranslated(pd.Mass()[0], pd.Mass()[1], pd.Mass()[2]);
	RenderBall(0.7f, 0.7f, 0.7f, 2);
	glPopMatrix();

	for (int i=0;i<pd.TunnelMass().size();++i)
	{
		Point & pt = pd.TunnelMass(i);
		glPushMatrix(); 
		glTranslatef((float)pt[0], (float)pt[1], (float)pt[2]);
		RenderBall(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2], 1.3f);
		glPopMatrix(); 
	}

	if (pd.AutoSurgeryPts().size() && show_sharpvert)
	{
		for (int i=pd.Genus(); i<pd.TunnelMass().size(); ++i)
		{
			std::pair<Vertex *, Vertex *> &cpair = pd.AutoSurgeryPts()[i-pd.Genus()];
			Point &pt1 = cpair.first->point();
			glPushMatrix(); 
			glTranslatef((float)pt1[0], (float)pt1[1], (float)pt1[2]);
			RenderBall(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2], 0.4f);
			glPopMatrix(); 

			Point &pt2 = cpair.second->point();
			glPushMatrix(); 
			glTranslatef((float)pt2[0], (float)pt2[1], (float)pt2[2]);
			RenderBall(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2], 0.4f);
			glPopMatrix(); 
		}
	}

	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(material_mode);

	///////////////////////////////////////////////////
	//render transparent surfaces
	///////////////////////////////////////////////////
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);

	glEnable(GL_CULL_FACE);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	Mesh * omesh = pd.CMesh();

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glPolygonMode(GL_FRONT, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_LINE:
		glPolygonMode(GL_FRONT, GL_LINE);
		glEnable(GL_POLYGON_OFFSET_LINE);
		glPolygonOffset(1.0, 1.0);
		break;
	case POLYGON_POINT:
		glPolygonMode(GL_FRONT, GL_POINT);
		glEnable(GL_POLYGON_OFFSET_POINT);
		glPolygonOffset(1.0, 1.0);
		glPointSize(3.0);
		break;
	}
		
	if (!omesh->hasNormal())
	{
		GetNormals::GetNorm(omesh, false);
	}
				
	glBegin(GL_TRIANGLES);

	Vector3f cColor(0.0f, 0.2f, 0.7f);		
	glColor4f(cColor[0], cColor[1], cColor[2], 0.4f);

	for(MeshFaceIterator fiter(omesh); !fiter.end(); ++fiter)
	{
		Face *f = *fiter;

		Vertex * v[3];
		HalfEdge * he = f->halfedge();
		v[0] = he->source();
		v[1] = he->target();
		v[2] = he->he_next()->target();

		Point * p[3];
		Point * n[3];

		p[0] = &v[0]->point();
		p[1] = &v[1]->point();
		p[2] = &v[2]->point();

		n[0] = &omesh->verNormal(v[0]);
		n[1] = &omesh->verNormal(v[1]);
		n[2] = &omesh->verNormal(v[2]);

		// sharp face
		if (show_sharpface && f->sharp())
		{
			DoDisableLights();
			glColor4f(sharpFaceColor[0], sharpFaceColor[1], sharpFaceColor[2], 0.9f);
		}


		for (int i=0;i<3;++i)
		{
			glNormal3d( (*n[i])[0], (*n[i])[1], (*n[i])[2] );
			glVertex3d( (*p[i])[0], (*p[i])[1], (*p[i])[2] );
		}

		if (show_sharpface && f->sharp())
		{
			if (enable_lighting)
			{
				DoEnableLights();
			}
			glColor4f(cColor[0], cColor[1], cColor[2], 0.4f);
		}

		
	}
		
	glEnd();

	glPolygonMode(GL_FRONT, GL_FILL);

	switch(polygon_mode)
	{
	case POLYGON_FILL:
		glDisable(GL_POLYGON_OFFSET_FILL);
		break;
	case POLYGON_LINE:
		glDisable(GL_POLYGON_OFFSET_LINE);
		break;
	case POLYGON_POINT:
		glDisable(GL_POLYGON_OFFSET_POINT);
		glPointSize(1.0);
		break;
	}

	glDisable (GL_BLEND);
	glDisable(GL_CULL_FACE);

	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(material_mode);

}


void PDecompRenderer::RenderGuideCurve()
{
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);	

	// Draw the hidden lines (edges)
	DoDisableLights();

	// anti-aliasing
	glEnable (GL_LINE_SMOOTH);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);

	glLineWidth(2.0);
	
	glBegin(GL_LINES);
	assert(pd->VGenus() <= 9);
		
	for (int i=0;i<guideEdges.size();++i)
	{
		Curve * c = guideEdges[i];
		if (!c)
			continue;
		glColor3f(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2]);			
		for (int j=0;j<c->heSize();++j)
		{
			HalfEdge * he = (*c)[j];
			if (!renderingMesh->hasHalfEdge(he))
				break;
			Point & p1 = he->source()->point();
			Point & p2 = he->target()->point();
			glVertex3f(p1[0], p1[1], p1[2]);
			glVertex3f(p2[0], p2[1], p2[2]);	
		}
	}

	//std::cout << pd->GuideCurvesList().size() << std::endl;

	for (int i=0;i<pd->GuideCurvesList().size();++i)
	{
		Curve * c = pd->GuideCurvesList()[i];
		if (!c)
			continue;
		glColor3f(HomoRGB[i][0], HomoRGB[i][1], HomoRGB[i][2]);			
		for (int j=0;j<c->heSize();++j)
		{
			HalfEdge * he = (*c)[j];
			if (!renderingMesh->hasHalfEdge(he))
			{
				assert(0);
				break;
			}
			Point & p1 = he->source()->point();
			Point & p2 = he->target()->point();
			glVertex3f(p1[0], p1[1], p1[2]);
			glVertex3f(p2[0], p2[1], p2[2]);	
		}
	}
				
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glDisable (GL_LINE_SMOOTH);
	glDisable (GL_BLEND);

	glDisable(GL_COLOR_MATERIAL);
}

void PDecompRenderer::RenderModel4Picking(Mesh *pMesh, int pmode)
{
	if (!pMesh) return;
	
	bool flag = false;

	MeshFaceIterator fiter(pMesh);
	MeshVertexIterator viter(pMesh);
	MeshEdgeIterator eiter(pMesh);

	switch (pmode)
	{
	case PICKING_FACE:
		for(; !fiter.end(); ++fiter)
		{
			Face *f = *fiter;
			glLoadName(f->id());
			glBegin(GL_TRIANGLES);
			for(FaceVertexIterator fviter(f); !fviter.end(); ++fviter)
			{
				Vertex *v = *fviter;
				Point & p = v->point();

				glVertex3f(p[0], p[1], p[2]);
			}
			glEnd();
		}
		break;
	case PICKING_VERTEX:
		for(; !viter.end(); ++viter)
		{
			Vertex *v = *viter;
			glLoadName(v->id());
			Point &p = v->point();			
			glBegin(GL_POINTS);
			glVertex3f( p[0], p[1], p[2] );
			glEnd();
		}
		break;
	case PICKING_EDGE:
		for(; !eiter.end(); ++eiter )
		{
			Edge * e = *eiter;
			Vertex * v1 = pMesh->edgeVertex1(e);
			Vertex * v2 = pMesh->edgeVertex2(e);

			int vid1 = e->vid(0); //e->vertex(0);
			int vid2 = e->vid(1); //vertex(1);
			// Note, the order of pushing names onto stack is important!
			if (!flag)
			{
				glPopName();
				glPushName(vid2);
				glPushName(vid1);
				flag = true;
			}
			else
			{
				glPopName(); 
				glPopName();
				glPushName(vid2);
				glPushName(vid1);
			}
			

			Point & p1 = v1->point();
			Point & p2 = v2->point();

			glBegin(GL_LINES);
			glVertex3f(p1[0], p1[1], p1[2]);
			glVertex3f(p2[0], p2[1], p2[2]);
			glEnd();
		}		
		break;
	default:
		break;
	}
}




void PDecompRenderer::RenderNormal(Mesh *pMesh)
{
	if (!pMesh) return;

	DoDisableLights();


	glColor3d(normalColor[0], normalColor[1], normalColor[2]);

	for (MeshVertexIterator viter(pMesh); !viter.end(); ++viter)
	{
		Vertex *v = *viter;

		double n[3];

		Point &p = v->point();			
		TraitParser::ReadTrait(v->string(), "normal", 3, n);


		glBegin(GL_LINES);
			glVertex3d(p[0], p[1], p[2]);
			glVertex3d(p[0]+n[0]*.1, p[1]+n[1]*.1, p[2]+n[2]*.1);
		glEnd();
	}

	if (enable_lighting)
	{
		DoEnableLights();
	}

}

void PDecompRenderer::RenderModelColorCode(Mesh *pMesh)
{
}


void PDecompRenderer::RenderBoundingBox(int mInd)
{
	vec3d blb, trf, centroid;

	blb = m_bbox[mInd].GetCorner1();
	trf = m_bbox[mInd].GetCorner2();
	centroid = m_bbox[mInd].center();

	DoDisableLights();
	glColor3f(0.6f, 0.6f, 0.6f);

	glLineWidth(2.0);
	glBegin(GL_LINES);
		glVertex3d(blb.x, blb.y, blb.z);	glVertex3d(trf.x, blb.y, blb.z);
		glVertex3d(blb.x, blb.y, trf.z);	glVertex3d(trf.x, blb.y, trf.z);
		glVertex3d(blb.x, trf.y, trf.z);	glVertex3d(trf.x, trf.y, trf.z);
		glVertex3d(blb.x, trf.y, blb.z);	glVertex3d(trf.x, trf.y, blb.z);

		glVertex3d(blb.x, blb.y, blb.z);	glVertex3d(blb.x, blb.y, trf.z);
		glVertex3d(blb.x, trf.y, blb.z);	glVertex3d(blb.x, trf.y, trf.z);
		glVertex3d(trf.x, trf.y, blb.z);	glVertex3d(trf.x, trf.y, trf.z);
		glVertex3d(trf.x, blb.y, blb.z);	glVertex3d(trf.x, blb.y, trf.z);

		glVertex3d(blb.x, blb.y, blb.z);	glVertex3d(blb.x, trf.y, blb.z);
		glVertex3d(trf.x, blb.y, blb.z);	glVertex3d(trf.x, trf.y, blb.z);
		glVertex3d(trf.x, blb.y, trf.z);	glVertex3d(trf.x, trf.y, trf.z);
		glVertex3d(blb.x, blb.y, trf.z);	glVertex3d(blb.x, trf.y, trf.z);
	glEnd();
	glLineWidth(1.0);

	if (enable_lighting)
	{
		DoEnableLights();
	}

	glPushMatrix();
	glTranslated(blb.x, blb.y, blb.z);
	RenderCoordinate();
	glPopMatrix();

	glPushMatrix();
	glTranslated(centroid.x, centroid.y, centroid.z);
	RenderBall(1.0f, 1.0f, 1.0f, 2);
	glPopMatrix();

}



void PDecompRenderer::RenderFloor(int mInd)
{
	const double INFINITY = 100;
	double y;

	y = m_bbox[0].minY();

	DoDisableLights();

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture2_name[2]);

	glColor3f(1.0f, 1.0f, 1.0f);

	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3d(-INFINITY, y, INFINITY);
		glTexCoord2f(0.0f, 64.0f);	
		glVertex3d(INFINITY, y, INFINITY);
		glTexCoord2f(64.0f, 64.0f);
		glVertex3d(INFINITY, y, -INFINITY);
		glTexCoord2f(64.0f, 0.0f);
		glVertex3d(-INFINITY, y, -INFINITY);
	glEnd();


    glDisable(GL_TEXTURE_2D);
	if (enable_lighting)
	{
		DoEnableLights();
	}
}


void PDecompRenderer::RenderTextOnPoint(const char text[], Point & pos)
{	
	// raster fonts
	DoDisableLights();

	glLineWidth(1.0);

	glColor3f(0.0f, 0.0f, 0.0f);
	glPushMatrix();
	glTranslatef(pos[0], pos[1], pos[2]);
	glScalef(0.1f, 0.1f, 0.1f);
	RenderText(text, GLUT_STROKE_MONO_ROMAN, FONT_STROKE);	
	glPopMatrix();
	glFlush();

	if (enable_lighting)
	{
		DoEnableLights();
	}
}


void PDecompRenderer::RenderBall(int mat_mode, float size)
{
	int cur_mat = material_mode;
	SetMaterial(mat_mode);

	glShadeModel(GL_SMOOTH);
	
	glPushMatrix();
		glScalef(size, size, size);
		glCallList(quadric_list);
	glPopMatrix();

	if (shading_mode == SHADING_SMOOTH)
		glShadeModel(GL_SMOOTH);
	else 
		glShadeModel(GL_FLAT);

	SetMaterial(cur_mat);
}


void PDecompRenderer::RenderBall(float R, float G, float B, float size)
{
	int cur_mat = material_mode;
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_SPECULAR);
	glColor3f(0.7f, 0.7f, 0.7f);
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(0.3f, 0.3f, 0.3f);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glColor3f(R, G, B);

	glShadeModel(GL_SMOOTH);

	glPushMatrix();
		glScalef(size, size, size);
		glCallList(quadric_list);
	glPopMatrix();

	if (shading_mode == SHADING_SMOOTH)
		glShadeModel(GL_SMOOTH);
	else 
		glShadeModel(GL_FLAT);

	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(cur_mat);
}


void PDecompRenderer::RenderArrow(float R, float G, float B, float size)
{
	int cur_mat = material_mode;
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT, GL_SPECULAR);
	glColor3f(0.7f, 0.7f, 0.7f);
	glColorMaterial(GL_FRONT, GL_AMBIENT);
	glColor3f(0.3f, 0.3f, 0.3f);
	glColorMaterial(GL_FRONT, GL_DIFFUSE);
	glColor3f(R, G, B);
	
	glPushMatrix();
		glScalef(size, size, size);
		glPushMatrix();
		glTranslatef(0.0f,0.0f, 0.4f);
		glCallList(quadric_list+1);		
		glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
		glCallList(quadric_list+2);
		glPopMatrix();

		glCallList(quadric_list+3);
		glScalef(0.5f, 0.5f, 0.5f);
		glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
		glCallList(quadric_list+2);
	glPopMatrix();

	glDisable(GL_COLOR_MATERIAL);
	SetMaterial(cur_mat);
}


void PDecompRenderer::RenderCoordinate()
{
	glPushMatrix();
		glPushMatrix();
		glRotated(90, 0, 1, 0);
		RenderArrow(0.8f, 0.0f, 0.0f);
		glPopMatrix();

		glPushMatrix();
		glRotated(-90, 1, 0, 0);
		RenderArrow(0.0f, 0.8f, 0.0f);
		glPopMatrix();

		RenderArrow(0.0f, 0.0f, 0.8f);


	glPopMatrix();
}

void PDecompRenderer::RenderText(const char *text, void *font, int font_flag)
{
	if (!text)
		return;

	int nchar = strlen(text);

	int i;
	if (font_flag == FONT_STROKE)
	{
		for (i = 0; i < nchar; ++i)
			glutStrokeCharacter(font, text[i]);
	}
	else
	{
		for (i = 0; i < nchar; ++i)
			glutBitmapCharacter(font, text[i]);
	}
}


void PDecompRenderer::MouseDown(int button, int x, int y)
{
	float nx, ny;
	Mousei2f(x, y, &nx, &ny);
		
	last_mouse_x = nx;
	last_mouse_y = ny;
	last_button = button;

	switch (button) 
	{
	case FL_LEFT_MOUSE:
		if (Fl::event_shift())
		{
			theCamera.status = Camera::ZOOMING;
			cursor(FL_CURSOR_NS);
		}
		else if (Fl::event_ctrl())
		{
			theCamera.status = Camera::PANNING;
			cursor(FL_CURSOR_WE);
		}
		else if (Fl::event_alt())
		{
			theCamera.status = Camera::ROTATING;
			cursor(FL_CURSOR_NESW);
		}
		else
			theCamera.status = Camera::IDLING;
        

		if (enable_picking)
		{
			DoPicking(x, y, picking_mode);
			cursor(FL_CURSOR_CROSS);
		}

		break;
    case FL_MIDDLE_MOUSE:
		break;
    case FL_RIGHT_MOUSE:
		break;
    default:
        break;
	}

	
}


void PDecompRenderer::MouseUp(int button, int x, int y)
{
}


void PDecompRenderer::MouseMotion(int button, int x, int y)
{
	float cur_mouse_x, cur_mouse_y;
	float delta_x, delta_y;

	Mousei2f(x, y, &cur_mouse_x, &cur_mouse_y);

	delta_x = cur_mouse_x - last_mouse_x;
	delta_y = cur_mouse_y - last_mouse_y;

	if (last_button == FL_LEFT_MOUSE)
	{
		switch (theCamera.status)
		{
		case Camera::PANNING:
			Translate(delta_x, delta_y, 0);
			break;
		case Camera::ROTATING:
			Rotate(last_mouse_x, last_mouse_y, cur_mouse_x, cur_mouse_y);
			break;
		case Camera::ZOOMING:
			Translate(0, 0, delta_y);
			break;
		case Camera::IDLING:
		default:
			break;
		}
	}

	last_mouse_x = cur_mouse_x;
	last_mouse_y = cur_mouse_y;

	redraw();
}

void PDecompRenderer::ClearPicks()
{
	selectedVertices.clear();
	cSelectedEdges.clear();
	for (int i=0;i<guideEdges.size();++i)
	{
		Curve * c = guideEdges[i];
		delete c;
	}
	guideEdges.clear();
}

void PDecompRenderer::DoPicking(int x, int y, int pmode)
{
#define DEBUG_DOPICKING	0
	GLint hits;
	GLint viewport[4];
	GLuint selectBuf[SELBUF_SIZE];
	glGetIntegerv(GL_VIEWPORT, viewport);

	glSelectBuffer(SELBUF_SIZE, selectBuf);
	glRenderMode(GL_SELECT);

	glInitNames();
	glPushName(0);

	glPushMatrix();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPickMatrix((GLdouble)x, (GLdouble)(viewport[3] - y), 5, 5, viewport);
	theCamera.Projection(view_dist/DOF, view_dist*DOF, win_width, win_height);
	glMatrixMode(GL_MODELVIEW);
	theCamera.Modelview();

	// currently only picking on point set supported
	RenderModel4Picking(renderingMesh, pmode);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glFlush();

	hits = glRenderMode(GL_RENDER);
	if (hits < 0)
	{
		fprintf(stderr, "Error: Selection Buffer Overflow!\n");
		return;
	}

#if DEBUG_DOPICKING
	printf("hits: %d\n", hits);
#endif

	ProcessHits(hits, selectBuf, pmode);
	redraw();
#undef DEBUG_DOPICKING
}




void PDecompRenderer::ProcessHits(GLint hits, GLuint buffer[], int pmode)
{

	int i, j;
	GLuint names, *ptr;
	vector<pair<GLfloat, GLuint> > name_array;
	vector<pair<GLfloat, GLuint> >::iterator niter;
	vector<pair<GLfloat, pair<GLuint, GLuint> > > edge_array;
	vector<pair<GLfloat, pair<GLuint, GLuint> > >::iterator eiter;
	
	pair<GLfloat, GLuint> name_pair;
	pair<GLfloat, GLuint> other_pair;
	pair<GLfloat, pair<GLuint, GLuint> > edge_pair;

	ptr = (GLuint *) buffer;
	
	for (i = 0; i < hits; ++i)
	{
		names = *ptr;
		ptr += 3;

		for (j = 0; j < names; ++j)
		{
			// In vertex picking mode, we will have to eliminate duplicates
			// Duplicates don't happen in face/edge mode
			// This is related to the way we render the model for picking
			name_pair = std::make_pair((GLfloat)*(ptr-2)/0x7fffffff, (*ptr));
			name_array.push_back(name_pair);

#if DEBUG_PROCESSHITS
			fprintf(stderr, "name: %u z1: %g z2: %g\n", *ptr, (float)*(ptr-2)/0x7fffffff, (float)*(ptr-1)/0x7fffffff);
#endif				
			ptr++;
		}
	}

	// A minor problem: if two points are collinear they may both be picked
	// we need to throw away the one that is farther away
	// A tricky problem: the point which we don't actually see (hidden further away)
	// may gets picked if the picking is not accurate enough and does not actually 
	// pick the point in the front which we DO see
	// Solution: sort the names according to their z1 value
	switch(pmode)
	{
	case PICKING_FACE:
	case PICKING_VERTEX:
		//commented by xav
		if (name_array.size() > 1)
		{
			std::sort(name_array.begin(), name_array.end());
		}
		break;
	case PICKING_EDGE:
		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
		{
			name_pair = (*niter);
			++niter;
			other_pair = (*niter);
			
			GLfloat z1min = __min(name_pair.first, other_pair.first);
			edge_array.push_back(std::make_pair(z1min, std::make_pair(name_pair.second, other_pair.second)));
		}

		if (edge_array.size() > 1)
		{
			std::sort(edge_array.begin(), edge_array.end());
		}

		break;
	default:
		break;
	}


	if (name_array.empty())
	{
#if DEBUG_PROCESSHITS
		fprintf(stderr, "Name array empty!\n");
#endif
		return;
	}
		
	MeshFaceIterator fiter(renderingMesh);
	MeshVertexIterator viter(renderingMesh);

	switch (pmode)
	{
	// process each name
	case PICKING_FACE:		
		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
		{
			name_pair = *niter;
			Face *f = renderingMesh->idFace(name_pair.second);
			if (f)
			{
#if DEBUG_PROCESSHITS
				fprintf(stdout, "Face %d picked!\n", f->id());
#endif
				f->sharp() = !f->sharp();
			}
#if DEBUG_PROCESSHITS
			else
				fprintf(stderr, "Face not found!\n");
#endif
			break;	
		}
		break;
	case PICKING_VERTEX:
		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
		{
			name_pair = *niter;
			Vertex *v = renderingMesh->idVertex(name_pair.second);

			if (v)
			{
#if DEBUG_PROCESSHITS
				fprintf(stdout, "Vertex %d picked!\n", v->id());
#endif			
				v->sharp() = !v->sharp();
				selectedVertices.push_back(v);
			}
#if DEBUG_PROCESSHITS
			else
				fprintf(stderr, "Vertex not found!\n");
#endif
			break;
		}		
		break;
	case PICKING_EDGE:
		for (eiter = edge_array.begin(); eiter != edge_array.end(); ++eiter)
		{
			edge_pair = *eiter;
			Vertex *v0 = renderingMesh->idVertex(edge_pair.second.first);
			Vertex *v1 = renderingMesh->idVertex(edge_pair.second.second);
			assert(v0);
			assert(v1);
			Edge *e = renderingMesh->vertexEdge(v0, v1);
			if (e)
			{
#if DEBUG_PROCESSHITS
				fprintf(stdout, "Edge (%d,%d) picked!\n", e->vertex(0), e->vertex(1));
#endif
				//fprintf(stdout, "Edge (%d,%d) picked!\n", e->vid(0), e->vid(1));
				e->sharp() = !e->sharp();
				if (e->sharp())
				{
					cSelectedEdges.push_back(e);
				}
				else
				{
					std::vector<Edge *>::iterator eiter = std::find(cSelectedEdges.begin(),
						cSelectedEdges.end(), e);
					if (eiter!=cSelectedEdges.end())
						cSelectedEdges.erase(eiter);					
				}
			}
#if DEBUG_PROCESSHITS
			else
				fprintf(stderr, "Edge not found!\n");
#endif
			break;
		}		

		break;
	default:
		break;
	}
#undef DEBUG_PROCESSHITS

}


void PDecompRenderer::SetSurgeryPoint(int handleNum)
{
	surgeryHandle = handleNum;
}

void PDecompRenderer::SetGuideCurve(int handleNum)
{
	if (cSelectedEdges.size())
	{
		Curve * ncurve = new Curve;
		for (int i=0;i<cSelectedEdges.size();++i)
		{
			ncurve->AppendHalfEdge(cSelectedEdges[i]->halfedge(0));
		}
		if (!guideEdges.size())
			guideEdges.resize(pd->VGenus(), NULL);
		guideEdges[handleNum] = ncurve;	
		cSelectedEdges.clear();	
	}
}

void PDecompRenderer::SetBMarker(int mInd)
{
	if (show_psetseparate)
	{
		Pant * cpant = m_pantsSet->IDPant(show_pantInd);
		if (selectedVertices.size())
		{
			cpant->Marker(mInd) = selectedVertices[0];
			std::cout << "Mark vertex: " << selectedVertices[0]->id() << std::endl;
			selectedVertices.clear();
		}
	}
}

int PDecompRenderer::ExportTexture2D(const char *filename, GLubyte texture[TEX2_HEIGHT][TEX2_WIDTH][4])
{
	char seps[] = ".";
	int len = strlen(filename);
	const char *cp = filename, *ext = NULL;
	while ((cp = strstr(cp, seps)) != NULL)
	{
		ext = cp;
		cp++;
	}

	if (!ext || (strcmp(ext, ".bmp") && strcmp(ext, ".BMP")))
	{
			cerr << "Incorrect file extension " << ext << endl;
			return 0;
	}

	BitMap m_bitmap;
	GLubyte r, g, b;
				
	m_bitmap.CreateDirect(TEX2_WIDTH, TEX2_HEIGHT);
	for (int i = 0; i < TEX2_HEIGHT; ++i)
		for (int j = 0; j < TEX2_WIDTH; ++j)
		{
			r = texture[i][j][0];
			g = texture[i][j][1];
			b = texture[i][j][2];
			m_bitmap.SetPixel(j, TEX2_HEIGHT-i-1, RGB(r,g,b));
		}
			
	m_bitmap.Save(filename);

	return 1;
}



int PDecompRenderer::LoadUVTexture2D(const char * filename)
{
	if (m_ctexture2D.LoadFile(filename))
		return 0;
	else
		return 1;
	glBindTexture(GL_TEXTURE_2D, m_ctexture2D.TexName()); 
}


//int PDecompRenderer::LoadEnvmapTexture2D(const char *filename)
//{
//	BitMap m_bitmap;
//	m_bitmap.Load(filename);
//	if (!m_bitmap.IsLoaded())
//	{
//		cerr << "Warning: cannot load environment map texture data!" << endl;
//		return 1;
//	}
//
//	for (int i = 0; i < TEX2_HEIGHT; ++i)
//		for (int j = 0; j < TEX2_WIDTH; ++j)
//		{
//			COLORREF rgb = m_bitmap.GetPixel(j, TEX2_HEIGHT-i-1);
//			env_texture2D[i][j][0] = (GLubyte)GetRValue(rgb);
//			env_texture2D[i][j][1] = (GLubyte)GetGValue(rgb);
//			env_texture2D[i][j][2] = (GLubyte)GetBValue(rgb);
//			env_texture2D[i][j][3] = (GLubyte)255;		
//		}
//
//	RebindTexture2D(1, env_texture2D);
//
//	return 0;
//}
//
//
//int PDecompRenderer::LoadFloorTexture2D(const char * filename)
//{
//	BitMap m_bitmap;
//	m_bitmap.Load(filename);
//	if (!m_bitmap.IsLoaded())
//	{
//		cerr << "Warning: cannot load floor texture data!" << endl;
//		return 1;
//	}
//
//	for (int i = 0; i < TEX2_HEIGHT; ++i)
//		for (int j = 0; j < TEX2_WIDTH; ++j)
//		{
//			COLORREF rgb = m_bitmap.GetPixel(j, TEX2_HEIGHT-i-1);
//			floor_texture2D[i][j][0] = (GLubyte)GetRValue(rgb);
//			floor_texture2D[i][j][1] = (GLubyte)GetGValue(rgb);
//			floor_texture2D[i][j][2] = (GLubyte)GetBValue(rgb);
//			floor_texture2D[i][j][3] = (GLubyte)255;		
//		}
//
//	RebindTexture2D(2, floor_texture2D);
//
//	return 0;
//}
//
//
//
//void PDecompRenderer::RebindTexture2D(int tex_id, GLubyte texture[TEX2_HEIGHT][TEX2_WIDTH][4])
//{
//	assert(tex_id >=0 && tex_id < TEX2_COUNT);
//
//	glBindTexture(GL_TEXTURE_2D, texture_name[tex_id]);
//
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
//
//	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
//		0, GL_RGBA, GL_UNSIGNED_BYTE, texture);
//
//	if (gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, TEX2_WIDTH, TEX2_HEIGHT,
//		GL_RGBA, GL_UNSIGNED_BYTE, texture) != 0)
//		cerr << "Automatic 2D mipmap construction failed!" << endl;
//}



void PDecompRenderer::RandomColor(GLfloat RGB[3])
{	
	//srand((unsigned)1(NULL));
	for (int i = 0; i < 3; ++i)
		RGB[i] = (GLfloat)rand() / (GLfloat)RAND_MAX;
}



void PDecompRenderer::FixedColor(GLfloat RGB[3], int cluster_id)
{
	srand(1);
	srand(cluster_id*3 + 2);
	if (cluster_id % 2)
	{
		for (int i = 0; i < 3; ++i)
			RGB[i] = (GLfloat)rand() / (GLfloat)RAND_MAX;
	}
	else
	{
		for (int i = 2; i >= 0; i--)
			RGB[i] = (GLfloat)rand() / (GLfloat)RAND_MAX;
	}
}



void PDecompRenderer::CaptureScreen(const char * filename)
{
	BitMap m_bitmap;
	GLubyte r, g, b;

	GLubyte *pixel_buf = new GLubyte[3*win_width*win_height];

	if (!pixel_buf)
	{
		cerr << "Error: Not enough memeory!" << endl;
		exit(EXIT_FAILURE);
	}
			
	m_bitmap.CreateDirect(win_width, win_height);

	glReadBuffer(GL_FRONT);
	glReadPixels(0,0,win_width,win_height,GL_RGB, GL_UNSIGNED_BYTE, pixel_buf);
	glReadBuffer(GL_BACK);
	assert(pixel_buf);

	for (int i = 0; i < win_height; ++i)
		for (int j = 0; j < win_width; ++j)
		{
			r = *(pixel_buf+i*win_width*3+j*3);
			g = *(pixel_buf+i*win_width*3+j*3+1);
			b = *(pixel_buf+i*win_width*3+j*3+2);
			m_bitmap.SetPixel(j, win_height-i-1, RGB(r,g,b));
		}
	
	m_bitmap.Save(filename);
		
	delete [] pixel_buf;
}



void PDecompRenderer::BeginVideoCapture(const char *filename, int fps)
{
	//m_avigen = new CAVIGenerator;	
	////m_avigen->SetFileName(_T(filename));
	//m_avigen->SetFileName(LPCWSTR (filename)); // by xav
	//m_avigen->SetRate(fps);
	//m_avigen->SetBitmapHeader(win_width, win_height);
	//HRESULT hr = m_avigen->InitEngine();

	//if (FAILED(hr))
	//{
	//	cerr << (m_avigen->GetLastErrorMessage()) << endl;
	//	m_avigen->ReleaseEngine();
	//	delete m_avigen;
	//	m_avigen = NULL;
	//}
	//else
	//	m_recording = true;
}



void PDecompRenderer::CaptureVideoFrame()
{
	//assert(m_avigen);

	//if (!m_recording)
	//	return;

	//LPBITMAPINFOHEADER lpbih = m_avigen->GetBitmapHeader();
	//GLubyte *pixel_buf = new GLubyte[3*lpbih->biWidth* lpbih->biHeight];
	//if (!pixel_buf)
	//{
	//	cerr << "Error: Not enough memeory!" << endl;
	//	exit(EXIT_FAILURE);
	//}


	//glReadBuffer(GL_FRONT);
	//glReadPixels(0,0,lpbih->biWidth,lpbih->biHeight,GL_BGR_EXT, GL_UNSIGNED_BYTE, pixel_buf);
	//glReadBuffer(GL_BACK);
	//
	//assert(pixel_buf);
	//	
	//// Adding frame and continue if OK
	//HRESULT hr = m_avigen->AddFrame(pixel_buf);
	//if (FAILED(hr))
	//{
	//	cerr << (m_avigen->GetLastErrorMessage()) << endl;
	//	m_avigen->ReleaseEngine();
	//	delete m_avigen;
	//	m_avigen = NULL;
	//}

	//delete [] pixel_buf;
	//
}


void PDecompRenderer::EndVideoCapture()
{
	//assert(m_avigen);

	//m_recording = false;
	//m_avigen->ReleaseEngine();
	//delete m_avigen;
	//m_avigen = NULL;

}


void PDecompRenderer::SetFrameRate(int fps)
{
	//assert(m_avigen);
	//m_avigen->SetRate(fps);
}


void PDecompRenderer::SetVideoName(const char * filename)
{
	//assert(m_avigen);
	//m_avigen->SetFileName(LPCWSTR(filename)); // by xav
	//std::string t1 = filename;
	//std::wstring t=t1;
	//m_avigen->SetFileName(filename); // by xav
}


void PDecompRenderer::BeginEnvMap()
{
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
}


void PDecompRenderer::EndEnvMap()
{
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
}


void PDecompRenderer::UV2TexUV(double uv0[2], double uv1[2], double uv2[2], double tuv0[2], double tuv1[2], double tuv2[2])
{
	//Usually this won't happen in global parameterization
	if (uv1[0] < 0.2 && uv0[0] > 0.8)
		tuv1[0] = uv1[0] + 1;
	else if (uv1[0] > 0.8 && uv0[0] < 0.2)
		tuv1[0] = uv1[0] - 1;
	
	if (uv2[0] < 0.2 && uv0[0] > 0.8)
		tuv2[0] = uv2[0] + 1;
	else if (uv2[0] > 0.8 && uv0[0] < 0.2)
		tuv2[0] = uv2[0] - 1;

	if (uv1[1] < 0.2 && uv0[1] > 0.8)
		tuv1[1] = uv1[1] + 1;
	else if (uv1[1] > 0.8 && uv0[1] < 0.2)
		tuv1[1] = uv1[1] - 1;

	if (uv2[1] < 0.2 && uv0[1] > 0.8)
		tuv2[1] = uv2[1] + 1;
	else if (uv2[1] > 0.8 && uv0[1] < 0.2)
		tuv2[1] = uv2[1] - 1;
}

GLfloat PDecompRenderer::R = 1.0f; 
GLfloat PDecompRenderer::G = 1.0f; 
GLfloat PDecompRenderer::B = 1.0f;

//void PDecompRenderer::ProcessHits(GLint hits, GLuint buffer[], int pmode)
//{
//#define DEBUG_PROCESSHITS 0
//
//	int i, j;
//	GLuint names, *ptr;
//	vector<pair<GLfloat, GLuint> > name_array;
//	vector<pair<GLfloat, GLuint> >::iterator niter;
//	vector<pair<GLfloat, pair<GLuint, GLuint> > > edge_array;
//	vector<pair<GLfloat, pair<GLuint, GLuint> > >::iterator eiter;
//	
//	pair<GLfloat, GLuint> name_pair;
//	pair<GLfloat, GLuint> other_pair;
//	pair<GLfloat, pair<GLuint, GLuint> > edge_pair;
//
//	ptr = (GLuint *) buffer;
//	
//	for (i = 0; i < hits; ++i)
//	{
//		names = *ptr;
//		ptr += 3;
//
//		for (j = 0; j < names; ++j)
//		{
//			// In vertex picking mode, we will have to eliminate duplicates
//			// Duplicates don't happen in face/edge mode
//			// This is related to the way we render the model for picking
//			name_pair = std::make_pair((GLfloat)*(ptr-2)/0x7fffffff, (*ptr));
//			name_array.push_back(name_pair);
//
//#if DEBUG_PROCESSHITS
//			fprintf(stderr, "name: %u z1: %g z2: %g\n", *ptr, (float)*(ptr-2)/0x7fffffff, (float)*(ptr-1)/0x7fffffff);
//#endif				
//			ptr++;
//		}
//	}
//
//	// A minor problem: if two points are collinear they may both be picked
//	// we need to throw away the one that is farther away
//	// A tricky problem: the point which we don't actually see (hidden further away)
//	// may gets picked if the picking is not accurate enough and does not actually 
//	// pick the point in the front which we DO see
//	// Solution: sort the names according to their z1 value
//	switch(pmode)
//	{
//	case PICKING_FACE:
//	case PICKING_VERTEX:
//		if (name_array.size() > 1)
//		{
//			std::sort(name_array.begin(), name_array.end());
//		}
//		break;
//	case PICKING_EDGE:
//		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
//		{
//			name_pair = (*niter);
//			++niter;
//			other_pair = (*niter);
//			
//			GLfloat z1min = __min(name_pair.first, other_pair.first);
//			edge_array.push_back(std::make_pair(z1min, std::make_pair(name_pair.second, other_pair.second)));
//		}
//
//		if (edge_array.size() > 1)
//		{
//			std::sort(edge_array.begin(), edge_array.end());
//		}
//
//		break;
//	default:
//		break;
//	}
//
//
//	if (name_array.empty())
//	{
//#if DEBUG_PROCESSHITS
//		fprintf(stderr, "Name array empty!\n");
//#endif
//		return;
//	}
//		
//	MeshFaceIterator fiter(mesh[cMeshInd]);
//	MeshVertexIterator viter(mesh[cMeshInd]);
//
//	switch (pmode)
//	{
//	// process each name
//	case PICKING_FACE:		
//		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
//		{
//			name_pair = *niter;
//			Face *f = mesh[cMeshInd]->idFace(name_pair.second);
//			if (f)
//			{
//#if DEBUG_PROCESSHITS
//				fprintf(stdout, "Face %d picked!\n", f->id());
//#endif
//				f->sharp() = !f->sharp();
//			}
//#if DEBUG_PROCESSHITS
//			else
//				fprintf(stderr, "Face not found!\n");
//#endif
//			break;	
//		}
//		break;
//	case PICKING_VERTEX:
//		for (niter = name_array.begin(); niter != name_array.end(); ++niter)
//		{
//			name_pair = *niter;
//			Vertex *v = mesh[cMeshInd]->idVertex(name_pair.second);
//
//			if (v)
//			{
//#if DEBUG_PROCESSHITS
//				fprintf(stdout, "Vertex %d picked!\n", v->id());
//#endif			
//				v->sharp() = !v->sharp();
//			}
//#if DEBUG_PROCESSHITS
//			else
//				fprintf(stderr, "Vertex not found!\n");
//#endif
//			break;
//		}		
//		break;
//	case PICKING_EDGE:
//		for (eiter = edge_array.begin(); eiter != edge_array.end(); ++eiter)
//		{
//			edge_pair = *eiter;
//			Vertex *v0 = mesh[cMeshInd]->idVertex(edge_pair.second.first);
//			Vertex *v1 = mesh[cMeshInd]->idVertex(edge_pair.second.second);
//			assert(v0);
//			assert(v1);
//			Edge *e = mesh[cMeshInd]->vertexEdge(v0, v1);
//			if (e)
//			{
//#if DEBUG_PROCESSHITS
//				fprintf(stdout, "Edge (%d,%d) picked!\n", e->vertex(0), e->vertex(1));
//#endif
//				//fprintf(stdout, "Edge (%d,%d) picked!\n", e->vid(0), e->vid(1));
//				e->sharp() = !e->sharp();
//			}
//#if DEBUG_PROCESSHITS
//			else
//				fprintf(stderr, "Edge not found!\n");
//#endif
//			break;
//		}		
//
//		break;
//	default:
//		break;
//	}
//#undef DEBUG_PROCESSHITS
//
//}