#include "Munkres.h"
#include <iostream>
#include <cassert>

using namespace XMeshLib;

#pragma warning(disable:4267)
#pragma warning(disable:4018)

Munkres::Munkres(int _n, int _m)
: n(_n), m(_m), step(1)
{
	if (n>m)
	{
		std::cout << "Warning: make n <= m to guarantee a stable result!" <<std::endl;
	}
	R_cover = new int [n];
	C_cover = new int [m];
	Mask = new int * [n];
	min_m_n = (m<n)?m:n;
	path1.resize(m+n); 
	path2.resize(m+n);

	for (int i=0;i<n;++i)
	{
		R_cover[i] = 0;
		Mask[i] = new int [m];		
		for (int j=0;j<m;++j)
			Mask[i][j] = 0;
	}
	for (int j=0;j<m;++j)
	{
		C_cover[j] = 0;		
	}
	
	for (int i = 0; i<min_m_n; ++i)	
		path1[i] = path2[i] = -1;
}

void Munkres::OutputMatrix()
{
	std::cout<< "The input matrix is: " << std::endl;
	std::cout << "==============================================================" << std::endl;
	for (int i=0;i<n;++i)
	{
		for (int j=0;j<m;++j)
		{
			std::cout << Costs[i][j] << " ";
		}
		std::cout<<std::endl;
	}
	std::cout << "==============================================================" << std::endl;
}

void Munkres::find_a_zero(int & row, int & col)
{
	row = -1;
	col = -1;
	int i=0;
	bool done = false;
	while (!done)
	{
		for (int j=0;j<m;++j)
		{
			if (!Costs[i][j] && !R_cover[i] && !C_cover[j])
			{
				row = i;
				col = j;				
				return;
			}
		}
		++i;
		if (i>=n)
			done = true;
	}
}

bool Munkres::star_in_row(int row)
{	
	for (int j=0; j<m; ++j)
	{
		if (Mask[row][j]==1)
			return true;
	}
	return false;
}

int Munkres::find_star_in_row(const int row)
{
	int col = -1;
	for (int j=0;j<m;++j)
	{
		if (Mask[row][j]==1)
			col = j;
	}
	return col;
}

int Munkres::find_star_in_col(const int col)
{
	int row = -1;
	for (int i=0;i<n;++i)
	{
		if (Mask[i][col]==1)
		{
			row =i;
			return row;
		}
	}
	return row;
}

int Munkres::find_prime_in_row(const int row)
{
	int col = -1;
	for (int j=0;j<m;++j)
	{
		if (Mask[row][j]==2)
		{
			col = j;
			return col;
		}
	}
	return col;
}

void Munkres::convert_path(int count)
{
	for (int i=0;i<=count;++i)
	{
		if (Mask[path1[i]][path2[i]]==1)
		{
			Mask[path1[i]][path2[i]]=0;		
		}
		else
		{
			Mask[path1[i]][path2[i]]=1;
		}
	}
}

void Munkres::clear_covers()
{
	for (int i=0;i<n;++i)	
		R_cover[i] = 0;	
	for (int j=0;j<m;++j)
		C_cover[j] = 0;
}

void Munkres::erase_primes()
{
	for (int i=0;i<n;++i)
	{
		for (int j=0; j<m; ++j)
		{
			if (Mask[i][j]==2)
			{
				Mask[i][j] = 0;
			}
		}
	}
}

double Munkres::find_smallest()
{
	double minval = 1e9;
	for (int i=0;i<n;++i)
		for (int j=0;j<m;++j)
		{
			if (!R_cover[i] && !C_cover[j]) 
			{
				if (minval > Costs[i][j])
					minval = Costs[i][j];
			}
		}
	return minval;
}

void Munkres::Step_1()
{
	assert(step==1);
	for (int i=0;i<n;++i)
	{
		double minval = Costs[i][0];
		for (int j=1; j<m; ++j)
		{
			if (minval > Costs[i][j])			
				minval = Costs[i][j];
		}
		for (int j=0;j<m; ++j)
			Costs[i][j] -= minval;
	}
	step = 2;
}

void Munkres::Step_2()
{
	assert(step==2);
	for (int i=0;i<n; ++i)
		for (int j=0;j<m;++j)
		{
			if (!Costs[i][j] && !C_cover[j] && !R_cover[i])
			{
				Mask[i][j] = 1;
				R_cover[i] = 1;
				C_cover[j] = 1;				
			}
		}
	for (int i=0; i<n; ++i)
		R_cover[i] = 0;
	for (int j=0; j<m; ++j)
		C_cover[j] = 0;

	step =3;
}

void Munkres::Step_3()
{
	assert(step==3);
	for (int i=0;i<n; ++i)
		for (int j=0;j<m;++j)
		{
			if (Mask[i][j] == 1)
			{
				C_cover[j] = 1;
			}
		}
	int count = 0;
	for (int j=0;j<m;++j)
		count += C_cover[j];
	if (count >= min_m_n)
		step = 7;
	else
		step = 4;
}

void Munkres::Step_4()
{
	assert(step==4);
	bool done = false;
	int row, col;
	while (!done)
	{
		find_a_zero(row, col);
		if (row == -1)
		{
			done = true;
			step = 6;
		}
		else
		{
			Mask[row][col] = 2;
			if (star_in_row(row))
			{
				col = find_star_in_row(row);
				R_cover[row] = 1;
				C_cover[col] = 0;
			}
			else
			{
				done = true;
				step = 5;
				Z0_r=row;
				Z0_c=col;
			}
		}
	}
}

void Munkres::Step_5()
{
	assert(step==5);
	int count = 0;
	path1[count] = Z0_r;
	path2[count] = Z0_c;
	bool done = false;
	while (!done)
	{
		int r = find_star_in_col(path2[count]);
		if (r>=0)
		{
			++count;
			path1[count] = r;
			path2[count] = path2[count-1];
		}
		else
			done = true;
		if(!done)
		{
			int c = find_prime_in_row(path1[count]);
			++count;
			path1[count] = path1[count-1];
			path2[count] = c;
		}
	}
	convert_path(count);
	clear_covers();
	erase_primes();
	step = 3;
}

void Munkres::Step_6()
{
	assert(step==6);
	double minval = find_smallest();
	for (int i=0; i<n;++i)
	{
		for (int j=0;j<m;++j)
		{
			if (R_cover[i]==1)
				Costs[i][j] += minval;
			if (C_cover[j]==0)
				Costs[i][j] -= minval;
		}
	}
	step = 4;
}

void Munkres::Process()
{
	bool done = false;
	step = 1;
	while (!done)
	{
		switch (step)
		{
		case 1:
			Step_1();
			break;
		case 2:
			Step_2();
			break;
		case 3:
			Step_3();
			break;
		case 4:
			Step_4();
			break;
		case 5:
			Step_5();
			break;
		case 6:
			Step_6();
			break;
		default:
			done = true;
		}
	}

	//std::cout << "The best matching is :" << std::endl;
	for (int i=0;i<n;++i)
	{
		for (int j=0;j<m;++j)
		{
			if (Mask[i][j]==1)
			{
				std::pair<int, int> cpair;
				cpair.first = i;
				cpair.second = j;
				OptMatch.push_back(cpair); 
			}
		}
	}


	//std::cout << "The best matching is :" << std::endl;
	//for (int i=0;i<n;++i)
	//{
	//	for (int j=0;j<m;++j)
	//	{
	//		if (Mask[i][j]==1)
	//		{
	//			std::cout << i << " " << j << std::endl;
	//		}
	//	}
	//}
}

void Munkres::VectorsDist2Matrix(std::vector<Point> vlist1, std::vector<Point> vlist2,
						int & row, int & col, double ** & outMatrix)
{
	row = vlist1.size();
	col = vlist2.size();
	outMatrix = new double * [row];	
	for (int i=0;i<row; ++i)
	{
		outMatrix[i] = new double [col];
		for (int j=0;j<col;++j)
		{
			outMatrix[i][j] = (vlist1[i]-vlist2[j]).norm();
			outMatrix[i][j] *= outMatrix[i][j];
			//outMatrix[i][j] += 1;
		}
	}
}

double Munkres::EstimateFinalCosts(double **mat)
{
	double totalCost = 0;	
	for (int i=0;i<OptMatch.size();++i)
	{
		totalCost += mat[OptMatch[i].first][OptMatch[i].second];
	}
	return totalCost;
}

void Munkres::FreeMatrix(double ** m, int row)
{
	for (int i=0;i<row; ++i)
		delete [] m[i];
	delete [] m;
}
