#include "Mesh.h"
#include <algorithm>
#include "iterators.h"
#include <iostream>
#include <sstream>
#include <string>

#pragma warning (disable : 4786)
#pragma warning (disable : 4996)
#pragma warning (disable : 4267)
#pragma warning (disable : 4018)

using namespace XMeshLib;

 Mesh::~Mesh()
{	
	std::list<Face *>::iterator fiter = this->m_faces.begin();
	for (; fiter!=m_faces.end(); ++fiter)
	{
		Face * f = *fiter;
		delete f;
	}
	std::list<Edge *>::iterator eiter = this->m_edges.begin();
	for (; eiter!=m_edges.end(); ++eiter)
	{
		Edge * e = *eiter;
		HalfEdge * he1 = e->halfedge(0);
		HalfEdge * he2 = e->halfedge(1);
		delete he1;
		if (he2)
			delete he2;
		delete e;
	}
	std::list<Vertex *>::iterator viter = this->m_verts.begin();
	for (; viter!=m_verts.end(); ++viter)
	{
		Vertex * v = *viter;
		delete v;
	}	
}

bool Mesh::read( const char inputFile[])
{
	FILE * fp = fopen( inputFile, "r" );
	if( !fp )
	{
		printf("Can't open file %s\n", inputFile );
		return false;
	}

	char line[1024];
	int id;
	this->maxid=-1;
	this->maxfid = -1;

	m_verts.clear();
	m_edges.clear();
	m_faces.clear();

	while( !feof(fp)  )
	{	
		fgets( line, 1024, fp );
		if( !strlen( line ) )
			continue;

		char * str = strtok( line, " \r\n");
		if (!str)
			continue;
		if( !strcmp(str, "Vertex" ) ) 
		{
			str = strtok(NULL," \r\n{");
			id = atoi( str );
			Point p;
			for( int i = 0 ; i < 3; i ++ )
			{
				str = strtok(NULL," \r\n{");
				p[i] = atof( str );
			}		
			Vertex * v  = createVertex( id );
			v->point() = p;
			v->id()    = id;

			str = strtok( NULL, "\r\n");

			if(!str) 
				continue;

			std::string s(str);
			int sp = s.find("{");
			int ep = s.find("}");

			if( sp >= 0 && ep >= 0 )
			{
				v->string() = s.substr( sp+1, ep-sp-1 );
			}
			continue;
		}
		
		if( !strcmp(str,"Face") ) 
		{
			str = strtok(NULL, " \r\n");
			if (!str)
				continue;
			id = atoi( str );
			int vid[3];
			for( int i = 0; i < 3; i ++ )
			{
				str = strtok(NULL," \r\n{");
				vid[i] = atoi(str);
			}
			Face * f = createFace( vid, id );
			if (!str)
				continue;
			str = strtok( NULL, "\r\n");
			if( !str || strlen( str ) == 0 ) continue;
			std::string s(str);
			int sp = s.find("{");
			int ep = s.find("}");
			if( sp >= 0 && ep >= 0 )
			{
				f->string() = s.substr( sp+1, ep-sp-1 );
			}
			continue;
		}

		//read in edge attributes
		if( !strcmp(str,"Edge") )
		{
			str = strtok(NULL, " \r\n");
			if (!str)
				continue;
			int id0 = atoi( str );
			str = strtok(NULL, " \r\n");
			int id1 = atoi( str );
			Edge * edge = idEdge( id0, id1 );
			if (!edge)
				continue;
			str = strtok(NULL, "\r\n");
			if (!str)
				continue;
			std::string s(str);
			int sp = s.find("{");
			int ep = s.find("}");
			if( sp >= 0 && ep >= 0 )
			{
				edge->string() = s.substr( sp+1, ep-sp-1 );
			}
			continue;
		}
	}
	
	//Label boundary edges
	for( std::list<Edge *>::iterator eiter = m_edges.begin();
		eiter!=m_edges.end(); ++eiter)
	{		
		Edge * edge = *eiter;
		HalfEdge * he[2];

		he[0] = edge->halfedge(0);
		he[1] = edge->halfedge(1);
		
		assert( he[0] );
		

		if( he[1] )
		{
			assert( he[0]->target() == he[1]->source() && he[0]->source() == he[1]->target() );

			if( he[0]->target()->id() < he[0]->source()->id() )
			{
				edge->halfedge(0) = he[1];
				edge->halfedge(1) = he[0];
			}

			assert( edgeVertex1(edge)->id() < edgeVertex2(edge)->id() );
		}
		else
		{
			he[0]->vertex()->boundary() = true;
			he[0]->he_prev()->vertex()->boundary()  = true;
		}
	}
	fclose(fp);
	return true;
}



bool Mesh::write( const char outputFile[] )
{
	FILE * fp = fopen( outputFile,"w");
	if ( !fp )
	{
		printf("Cannot open file %s to write\n", outputFile);
		return false;
	}

	std::list<Vertex *>::iterator vit;
	for (vit=m_verts.begin(); vit!=m_verts.end(); ++vit)
	{
		Vertex * ver = *vit;

		std::ostringstream oss;
		oss.precision(12);
		oss.setf(std::ios::fixed,std::ios::floatfield);
		oss << "Vertex " << ver->id() << " " << ver->point()[0] << " " << ver->point()[1] << " " << ver->point()[2];
		fprintf(fp, "%s ", oss.str().c_str());

		if (ver->string().size() > 0)
		{
			fprintf(fp, "{%s}", ver->string().c_str() );
		}
		fprintf(fp, "\n");
	}

	std::list<Face *>::iterator fit;
	for (fit=m_faces.begin(); fit!=m_faces.end(); ++fit)
	{
		Face * face = *fit;
		HalfEdge * the0 = face->halfedge();
		HalfEdge * the1 = the0->he_next();
		int v0 = the0->source()->id();
		int v1 = the0->target()->id();
		int v2 = the1->target()->id();
		fprintf(fp, "Face %d %d %d %d ", face->id(), v0, v1, v2);
		if (face->string().size() > 0)
		{
			fprintf(fp, "{%s}", face->string().c_str() );
		}
		fprintf(fp, "\n");
	}

	std::list<Edge *>::iterator eit;
	for (eit=m_edges.begin(); eit!=m_edges.end(); ++eit)
	{
		Edge * edge = *eit;		
		if (edge->string().size() > 0)
		{ //only output edges when they have special traits
			fprintf(fp, "Edge %d %d {%s} \n", edge->vid(0), edge->vid(1), edge->string().c_str() );
		}
	}
	fclose(fp);
	return true;
}

void Mesh::copyTo( Mesh & solid )
{
	for(std::list<Vertex *>::iterator viter = m_verts.begin();
		viter!=m_verts.end(); ++viter)
	{
		Vertex * v = *viter;
		Vertex * nv = solid.createVertex(v->id());
		nv->point() = v->point();
		nv->string() = v->string();		
		nv->boundary() = v->boundary();
	}
	solid.maxid = maxid;
	
	for(std::list<Face *>::iterator fiter = m_faces.begin();
		fiter!=m_faces.end(); ++fiter)
	{
		Face * f = *fiter;
		HalfEdge * he = f->halfedge();
		int v[3];
		for (int j=0; j<3; ++j)
		{
			v[j] = he->vertex()->id();
			he = he->he_next();
		}
		Face * nf = solid.createFace(v, f->id());
		nf->string() = f->string();
	}

	solid.id2Ver.clear(); 
	solid.id2Face.clear();
	
	for (MeshVertexIterator viter(&solid); !viter.end(); ++viter)
	{
		Vertex * v = *viter;
		solid.id2Ver[v->id()] = v;
	}
	for (MeshFaceIterator fiter(&solid); !fiter.end(); ++fiter)
	{
		Face * f = *fiter;
		solid.id2Face[f->id()] = f;
	}
}

Edge * Mesh::idEdge( int id0, int id1 )
{
	Vertex * v0 = idVertex(id0);
	Vertex * v1 = idVertex(id1);
	if (!v0 || !v1)
		return NULL;
	else
		return vertexEdge(v0, v1);
}

Edge * Mesh::vertexEdge( Vertex * v0, Vertex * v1 )
{

		std::list<Edge *>::iterator eit;
		for (eit = m_edges.begin(); 
			eit != m_edges.end(); 
			++eit)
		{
			Edge * te = *eit;
			HalfEdge * the0 = te->halfedge(0);
			Vertex * tv0 = the0->source();
			Vertex * tv1 = the0->target();		
			if (tv0 == v0 && tv1 == v1)
				return te;
			else if (tv0 == v1 && tv1 == v0)
				return te;
		}
		return NULL;
}

Face * Mesh::edgeFace1( Edge *  e )
{
	assert( e->halfedge(0) != NULL );
	return e->halfedge(0)->face();
}
Face * Mesh::edgeFace2( Edge *  e )
{
	assert( e->halfedge(1) != NULL );
	return e->halfedge(1)->face();
}

Face * Mesh::halfedgeFace( HalfEdge *  he )
{
	return he->face();
}

Face * Mesh::idFace(const int id )
{
	stdext::hash_map<int, Face *>::iterator cpos = id2Face.find(id);
	if (cpos == id2Face.end())
		return NULL;
	else
	{
		Face * f = cpos->second;
		assert(f->id()==id);
		return f;
	}
}

HalfEdge * Mesh::idHalfedge( int id0, int id1 )
{
	Edge * e = idEdge( id0, id1 );
	if (!e)
		return NULL;
	HalfEdge * he = e->halfedge(0);
	if( he->source()->id() == id0 && he->target()->id() == id1 ) 
		return he;
	he = e->halfedge(1);
	if (!he)
		return NULL;
	if( he->source()->id() == id0 && he->target()->id() == id1 ) 
		return he;
	assert(0);
	return NULL;
}

HalfEdge * Mesh::vertexHalfedge( Vertex * v0, Vertex * v1 )
{
	Edge * e = vertexEdge( v0, v1 );
	if (!e)
		return NULL;
	HalfEdge * he = e->halfedge(0);
	if( he->vertex() == v1 && he->he_prev()->vertex() == v0 ) 
		return he;
	he = e->halfedge(1);
	if (he)
	{
		assert( he->vertex() == v1 && he->he_prev()->vertex() == v0 );
	}
	return he;
}

Vertex * Mesh::edgeVertex1( Edge *  e )
{
	assert(e->halfedge(0));
	return e->halfedge(0)->source();
}

Vertex * Mesh::edgeVertex2( Edge *  e )
{
	assert(e->halfedge(0));
	return e->halfedge(0)->target();
}

Vertex * Mesh::idVertex( int id ) 
{
	stdext::hash_map<int, Vertex *>::iterator cpos = id2Ver.find(id);
	if (cpos == id2Ver.end())
		return NULL;
	else
	{
		Vertex * v = cpos->second;
		assert(v->id()==id);
		return v;
	}
}

Edge * Mesh::edgeFlip(Edge * e)
{
	HalfEdge * he1 = e->halfedge(0);
	HalfEdge * he2 = e->halfedge(1);
	if (!he1 || !he2)
	{
		// this edge is on boundary, can not do flip
		return NULL;
	}
	Face * f1 = he1->face();
	Face * f2 = he2->face();
	HalfEdge * he3 = he1->he_next();
	HalfEdge * he4 = he3->he_next();
	HalfEdge * he5 = he2->he_next();
	HalfEdge * he6 = he5->he_next();
	e->vid(0) = he3->target()->id();
	e->vid(1) = he5->target()->id();
	he1->target() = he3->target();
	he2->target() = he5->target();
	he4->he_next() = he5; he5->he_prev() = he4;
	he5->he_next() = he1; he1->he_prev() = he5;
	he1->he_next() = he4; he4->he_prev() = he1;
	he3->he_next() = he2; he2->he_prev() = he3;
	he2->he_next() = he6; he6->he_prev() = he2;
	he6->he_next() = he3; he3->he_prev() = he6;
	he5->face() = f1;
	he3->face() = f2;
	f1->halfedge() = he4;
	f2->halfedge() = he6;
	HalfEdge * h = e->halfedge(0);
	if( h->target()->id() < h->source()->id() )
	{
		e->halfedge(0) = e->halfedge(1);
		e->halfedge(1) = h;
	}
	he6->target()->halfedge() = he6;
	he4->target()->halfedge() = he4;
	return e;
}

Vertex * Mesh::faceSplit(Face * f, double bary[3])
{
	if (bary[0]<0 || bary[1]<0 || bary[2]<0)
	{
		// the split point is not inside the face, can't do split
		return NULL;
	}
	// first we create the new vertex
	HalfEdge * he[3];
	he[0] = f->halfedge();
	Vertex * v[4];	
	v[0] = he[0]->target();
	he[1] = he[0]->he_next();
	v[1] = he[1]->target();
	he[2] = he[1]->he_next();
	v[2] = he[2]->target();
	v[3] = createVertex( ++this->maxid );
	v[3]->point() = (v[0]->point() * bary[0] + v[1]->point() * bary[1] + v[2]->point() * bary[2]) / (bary[0] + bary[1] + bary[2]);

	int i,j;

	//delete the original big face	
	int maxfid = -1;
	for (std::list<Face *>::iterator fiter=m_faces.begin();
		fiter!=m_faces.end(); ++fiter)
	{		
		Face *f=*fiter;
		if (f->id() > maxfid)
			maxfid = f->id();
	}

	HalfEdge * hes[3][2];
	for(i = 0; i < 3; ++i)
	{
		for (j=0; j<2; ++j)
		{
			hes[i][j] = new HalfEdge;
			assert(hes[i][j]);
		}
		hes[i][0]->vertex() = v[3];
		v[3]->halfedge() = hes[i][0];
		hes[i][1]->vertex() = v[i];
		he[i]->he_next() = hes[i][0];  hes[i][0]->he_prev() = he[i];
	}

	//linking to each other for left 6 halfedges
	hes[0][0]->he_next() = hes[2][1];  hes[2][1]->he_prev() = hes[0][0];
	hes[1][0]->he_next() = hes[0][1];  hes[0][1]->he_prev() = hes[1][0];
	hes[2][0]->he_next() = hes[1][1];  hes[1][1]->he_prev() = hes[2][0];

	hes[2][1]->he_next() = he[0]; he[0]->he_prev() = hes[2][1];
	hes[0][1]->he_next() = he[1]; he[1]->he_prev() = hes[0][1];
	hes[1][1]->he_next() = he[2]; he[2]->he_prev() = hes[1][1];

	Face * f1 = createFace( ++maxfid );
	Face * f2 = createFace( ++maxfid );

	//linking to face
	hes[0][0]->face() = hes[2][1]->face() = he[0]->face() = f;
	hes[1][0]->face() = hes[0][1]->face() = he[1]->face() = f1;
	hes[2][0]->face() = hes[1][1]->face() = he[2]->face() = f2;
	f->halfedge() = he[0];
	f1->halfedge() = he[1];
	f2->halfedge() = he[2];

	//connecting with edge
	for( i = 0; i < 3; i ++ )
	{		
		int id0 = hes[i][0]->target()->id(); 
		int id1 = hes[i][1]->target()->id(); 
		Edge * e = new Edge(id0, id1);
		m_edges.push_back(e);
		hes[i][0]->edge() = hes[i][1]->edge() = e;
		e->halfedge(0) = hes[i][0];
		e->halfedge(1) = hes[i][1];
	}	

	return v[3];
}

Vertex * Mesh::EdgeSplit( Edge * e )
{

	// (1) Initialize: Get halfedge pointer
	HalfEdge * hes[2];
	hes[0] = e->halfedge(0);
	hes[1] = e->halfedge(1);
	int i,j,k;

	//(2) Remove edge
	std::list<Edge *>::iterator eiter = std::find(m_edges.begin(), m_edges.end(), e);
	if (eiter == m_edges.end())
	{
		std::cerr << "Error: Given edge does not exists in current mesh or is not valid!" << std::endl;
		return NULL;
	}
	//assert(eiter != m_edges.end());
	m_edges.erase(eiter);	


	//(3) Create a new vertex
	Vertex * nv = createVertex( ++maxid );	
	nv->boundary() = (!hes[1]);
	nv->point() = (hes[0]->source()->point() + hes[0]->target()->point())/2;
	
	//(4)
	std::list<Face *> new_faces;
	for( j = 0; j < 2; ++j )
	{
		Vertex * v[3];
		HalfEdge * he = hes[j];

		if( !he ) continue;

		//(4.1) Find and remove the face f connected to hes[j]
		Face * f = he->face();
		std::list<Face *>::iterator fiter = std::find(m_faces.begin(), m_faces.end(), f);
		assert(fiter!=m_faces.end());
		m_faces.erase(fiter);
		delete f;

		//(4.2) Remove the linkings from f's halfedges to their original connected edges
		for(i = 0; i < 3; ++i )
		{
			v[i] = he->target();			
			Edge * te = he->edge();
			if( te->halfedge(0) == he )
			{
				te->halfedge(0) = te->halfedge(1);
			}
			te->halfedge(1) = NULL;
			if (!te->halfedge(0))
			{//both halfedges have been removed
				//assert(te == e);
				if (te!=e)
				{
					std::list<Edge *>::iterator teiter = 
						std::find(m_edges.begin(), m_edges.end(), te);
					assert(teiter != m_edges.end());
					m_edges.erase(teiter);
					delete te;
				}
			}
			he = he->he_next();
		}

		//(4.3) Remove all halfedges in f
		for(k = 0; k < 3; k ++ )
		{
			HalfEdge * ne = he->he_next();
			delete he;
			he = ne;
		}
		hes[j] = NULL;

		//(4.4) Create new faces
		int vid[3];
		Vertex * w[3];
		w[0] = nv; w[1] = v[0]; w[2] = v[1];
		for(k = 0; k < 3; k ++ )
		{
			vid[k] = w[k]->id(); 
		}
		Face * nf = createFace( vid, ++maxfid );
		if (!nf)
		{
			std::cerr << "Error: creating face " << maxfid << " fails!" <<std::endl;
			return NULL;
		}
		new_faces.push_back( nf );

		w[0] = nv; w[1] = v[1]; w[2] = v[2];
		for(k = 0; k < 3; k ++ )
		{
			vid[k] = w[k]->id(); 
		}	
		nf = createFace( vid, ++maxfid );
		if (!nf)
		{
			std::cerr << "Error: creating face " << maxfid << " fails!" <<std::endl;
			return NULL;
		}
		new_faces.push_back( nf );
	}

	//(5) Link newly created halfedges and edges
	std::list<Face *>::iterator nfit = new_faces.begin();
	for(; nfit!=new_faces.end(); ++nfit)
	{
		Face * f = *nfit;
		HalfEdge * he = f->halfedge();

		do{
			Edge     * e = he->edge();
			if( e->halfedge(1) )
			{
				HalfEdge * h = e->halfedge(0);
				if( h->target()->id() < h->source()->id() )
				{
					e->halfedge(0) = e->halfedge(1);
					e->halfedge(1) = h;
				}
			}
			he = he->he_next();
		}while( he != f->halfedge() );
	}

	delete e;
	
	return nv;	
}



//create new geometric simplexes

Vertex * Mesh::createVertex( int id )
{
	Vertex * v = new Vertex;	
	v->id() = id;	
	if (id > this->maxid)
	{
		this->maxid = id;
	}	
	id2Ver[id] = v;
	m_verts.push_back( v );

	return v;
}

Edge * Mesh::createEdge( Vertex * start, Vertex * end )
{
	Edge * e = idEdge(start->id(), end->id());
	if (!e)
	{
		e = new Edge(start->id(), end->id());
		assert(e);
		m_edges.push_back( e );	
	}
	return e;
}

Face * Mesh::createFace( int id )
{
	Face * f = new Face();
	assert(f);
	f->id() = id;
	if (maxfid < id)
		maxfid = id;
	m_faces.push_back(f);	
	return f;
}

Face * Mesh::createFace( Vertex * ver[3] , int id )
{	
	stdext::hash_map<int, Face *>::iterator cpos = id2Face.find(id);
	if (cpos != id2Face.end())
	{//Return NULL if this id already exits
		std::cerr << "Try to duplicate create a face with existing id: " << id << " !" << std::endl;
		return NULL;
	}
	
	int i;
	Face * f = createFace( id );	
	id2Face[id] = f;
	//create halfedges
	HalfEdge * hes[3];
	for(i = 0; i < 3; i ++ )
	{
		hes[i] = new HalfEdge;
		assert( hes[i] );
		hes[i]->vertex() = ver[i];
		ver[i]->halfedge() = hes[i];
	}
	//linking to each other
	for( i = 0; i < 3; i ++ )
	{
		hes[i]->he_next() = hes[(i+1)%3];
		hes[i]->he_prev() = hes[(i+2)%3];
	}
	//linking to face
	for( i = 0; i < 3; i ++ )
	{
		hes[i]->face()   = f;
		f->halfedge()    = hes[i];
	}
	//connecting with edge
	for( i = 0; i < 3; i ++ )
	{
		//Edge * e = createEdge( ver[i], ver[(i+2)%3] );
		int id0 = ver[i]->id(); // target
		int id1 = ver[(i+2)%3]->id(); // source
		Edge * e;
		bool existE=false;
			Edge * tte = idEdge(id0, id1);
			if (tte)
			{
				existE = true;
				e = tte;
			}
			else 
				existE = false;
		if (!existE)
		{
			e = new Edge(id0, id1);
			m_edges.push_back(e);
		}
		if(!e->halfedge(0))
		{
			e->halfedge(0) = hes[i];
		}
		else
		{
			assert(!e->halfedge(1));
			e->halfedge(1) = hes[i];
		}
		hes[i]->edge() = e;
	}
	return f;
}

Face * Mesh::createFace( int v[3] , int id )
{
	Vertex * ver[3];
	for(int i = 0; i < 3; i ++ )
	{
		ver[i] =  idVertex( v[i] );
		if (!ver[i])
		{
			std::cerr << "Error: invalid vertex id: " << v[i] << " provided when creating face "
				<< id << " !" << std::endl;
			return NULL;
		}
	}
	Face * f = createFace(ver, id);
	return f;
}

HalfEdge * Mesh::vertexMostClwOutHalfEdge( Vertex *  v )
{
	return v->most_clw_out_halfedge();
};

HalfEdge * Mesh::vertexMostCcwOutHalfEdge( Vertex *  v )
{
	return v->most_ccw_out_halfedge();
};

HalfEdge * Mesh::vertexNextCcwOutHalfEdge( HalfEdge *  he )
{
	return he->ccw_rotate_about_source();
};

HalfEdge * Mesh::vertexNextClwOutHalfEdge( HalfEdge *  he )
{
	assert( he->he_sym() != NULL );
	return he->clw_rotate_about_source();
};

HalfEdge * Mesh::vertexMostClwInHalfEdge( Vertex *  v )
{
	return v->most_clw_in_halfedge();
};

HalfEdge * Mesh::vertexMostCcwInHalfEdge( Vertex *  v )
{
	return v->most_ccw_in_halfedge();
};

HalfEdge * Mesh::vertexNextCcwInHalfEdge( HalfEdge *  he )
{
	assert( he->he_sym() != NULL );
	return he->ccw_rotate_about_target();
};

HalfEdge * Mesh::vertexNextClwInHalfEdge( HalfEdge *  he )
{
	return he->clw_rotate_about_target();
};

HalfEdge * Mesh::faceNextClwHalfEdge( HalfEdge *  he )
{
	return he->he_prev();
};

HalfEdge * Mesh::faceNextCcwHalfEdge( HalfEdge *  he )
{
	return he->he_next();
};

HalfEdge * Mesh::faceMostCcwHalfEdge( Face *  face )
{
	return face->halfedge();
};

HalfEdge * Mesh::faceMostClwHalfEdge( Face *  face )
{
	return face->halfedge()->he_next();
};
