#ifndef _XMESHLIB_CORE_MESH_H_
#define _XMESHLIB_CORE_MESH_H_

#pragma warning (disable : 4786)
#pragma warning (disable : 4267)
#pragma warning (disable : 4018)

#include "Edge.h"
#include "Face.h"
#include "HalfEdge.h"
#include "Vertex.h"
#include "Point.h"
#include <hash_map>
#include <list>

#define SIMPLE


namespace XMeshLib{

	class Vertex;
	class HalfEdge;
	class Edge;
	class Face;

	class Mesh
	{
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//////////////								Methods										//////////////////
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	public:
		//(1) Constructor and Destructor
		Mesh()	
		{			
			maxid = maxfid = -1;			
		}
		~Mesh();

		//(2) I/O
		int numVertices() const {return m_verts.size();}										//number of vertices
		int numEdges() const {return m_edges.size();}										//number of edges
		int numFaces() const {return m_faces.size();}										//number of faces
		void copyTo( Mesh & targetMesh );			//copy current mesh to the target mesh 
		bool read( const char inFile[]);						//read a mesh from inFile
		bool write( const char outFile[]);					//write a mesh to outFile

		//(3) BASIC OPERATIONS
		//Check whether an element is on the boundary:
		bool isBoundary( Vertex *  v ) const {return v->boundary();}
		bool isBoundary( Edge *    e ) const {return 	((!(e->halfedge(0))||(!(e->halfedge(1))))?true:false);}
		bool isBoundary( HalfEdge *  he ) const {return !(he->he_sym());}

		//check whether the given vertex is on current mesh
		bool hasVertex(Vertex * ver) 
		{
			return (idVertex(ver->id())==ver);
		}
		bool hasHalfEdge(HalfEdge * he)
		{
			return (idVertex(he->vertex()->id())==he->vertex());
		}
		
		//indexing elements
		Edge *			idEdge( int id0, int id1 );
		Edge *			vertexEdge( Vertex * v0, Vertex * v1 );
		Face *			edgeFace1( Edge *  e );
		Face *			edgeFace2( Edge *  e );
		Face *			halfedgeFace( HalfEdge * he );
		Face *			idFace(const int id );		
		HalfEdge *		idHalfedge( int id0, int id1 );
		HalfEdge *		vertexHalfedge( Vertex * v0, Vertex * v1 );
		Vertex *			edgeVertex1( Edge *  e );
		Vertex *			edgeVertex2( Edge *  e );
		Vertex *			halfedgeVertex( HalfEdge * he ) {he->vertex();}		
		Vertex *			idVertex( int id );


		//(4) Mesh Modification Operators

		Vertex * createVertex(int id);										
		Edge * createEdge(Vertex * start, Vertex * end);
		Face * createFace(int id);
		Face * createFace(int v[], int id);
		Face * createFace(Vertex * v[], int id);

		/*
			Flip the input edge, and return it.
			if the flip works, return the resultant edge, otherwise return NULL
		*/
		Edge *			edgeFlip(Edge * e);

		// Split the given edge(input) into two. 
		//	  return :  newly created vertex or NULL(when the split fails);
		Vertex *			EdgeSplit(Edge * e);

		Vertex *			faceSplit(Face * f, double bary[3]);


		//(5) For Iterators
		HalfEdge * vertexMostClwOutHalfEdge( Vertex *  v );
		HalfEdge * vertexNextCcwOutHalfEdge( HalfEdge *  he );
		HalfEdge * vertexMostCcwOutHalfEdge( Vertex *  v );
		HalfEdge * vertexNextClwOutHalfEdge( HalfEdge *  he );
		HalfEdge * vertexMostClwInHalfEdge( Vertex *  v );
		HalfEdge * vertexNextCcwInHalfEdge( HalfEdge *  he );
		HalfEdge * vertexMostCcwInHalfEdge( Vertex *  v );
		HalfEdge * vertexNextClwInHalfEdge( HalfEdge *  he );
		HalfEdge * faceMostClwHalfEdge( Face *  f );
		HalfEdge * faceMostCcwHalfEdge( Face *  f );
		HalfEdge * faceNextCcwHalfEdge( HalfEdge *  he );
		HalfEdge * faceNextClwHalfEdge( HalfEdge *  he );	


		////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//////////////								Variables										//////////////////
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	
	protected:
		std::list<Edge *>				m_edges;		//edge container
		std::list<Vertex *>				m_verts;		//vertex container
		std::list<Face *>				m_faces;		//face container

		int maxid;
		int maxfid;
		stdext::hash_map<int, Vertex *>		id2Ver;			//given a vertex id, get its handle in the container
		stdext::hash_map<int, Face *>		id2Face;		//given a face id, get its handle in the container

	protected:
		friend class MeshVertexIterator;
		friend class MeshEdgeIterator;
		friend class MeshFaceIterator;
		friend class MeshHalfEdgeIterator;
		friend class MeshDelegate;

	};

}//name space XMeshLib

#endif //_XMESHLIB_CORE_MESH_H_ defined
